"""Module for the base distribution and the scipy distribution."""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Iterable, Sequence, Union, Optional

import numpy as np
import pandas as pd
import polars as pl


class BaseDistribution(ABC):
    """Abstract base class to define a distribution.

    All distributions should be derived from this class, and the following
    methods need to be implemented: _fit, draw, to_dict.
    """

    implements: str = "unknown"
    var_type: str = "unknown"
    provenance: str = "builtin"
    privacy: str = "none"
    is_unique: bool = False
    version: str = "1.0"

    @classmethod
    def fit(cls, series: Union[Sequence, pl.Series],
            *args, **kwargs) -> BaseDistribution:
        """Fit the distribution to the series.

        Parameters
        ----------
        series: pandas.Series
            Data to fit the distribution to.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        pd_series = cls._to_series(series)
        if len(pd_series) == 0:
            return cls.default_distribution()
        return cls._fit(pd_series, *args, **kwargs)

    @staticmethod
    def _to_series(values: Union[Sequence, pl.Series, pd.Series]):
        if isinstance(values, pl.Series):
            series = values.drop_nulls()
        elif isinstance(values, pd.Series):
            series = pl.Series(values).drop_nulls()  # pylint: disable=assignment-from-no-return
        else:
            series_array = np.array(values)
            series_array = series_array[~np.isnan(series_array)]
            series = pl.Series(series_array)
        return series

    @classmethod
    @abstractmethod
    def _fit(cls, values: pl.Series) -> BaseDistribution:
        """See fit method, but does not need to deal with NA's."""

    @abstractmethod
    def draw(self) -> object:
        """Draw a random element from the fitted distribution."""

    def draw_reset(self) -> None:
        """Reset the drawing of elements to start again."""

    @property
    def _params_formatted(self) -> str:
        return "\n".join(
            f"\t- {param}: {value}" for param,
            value in self._param_dict().items()
        )

    def __str__(self) -> str:
        """Return an easy to read formatted string for the distribution."""
        return (
            f"- Type: {self.implements}\n"
            f"- Provenance: {self.provenance}\n"
            f"- Parameters:\n"
            f"{self._params_formatted}\n"
        )

    @abstractmethod
    def _param_dict(self):
        """Get dictionary with the parameters of the distribution."""

    def to_dict(self) -> dict:
        """Convert the distribution to a dictionary."""
        return {
            "implements": self.implements,
            "version": self.version,
            "provenance": self.provenance,
            "class_name": self.__class__.__name__,
            "parameters": deepcopy(self._param_dict()),
        }

    @classmethod
    @abstractmethod
    def _param_schema(cls):
        """Get schema for the parameters of the distribution."""

    @classmethod
    def schema(cls) -> dict:
        """Create sub-schema to validate GMF file."""
        return {
            "type": "object",
            "properties": {
                "implements": {"const": cls.implements},
                "version": {"type": "string"},
                "provenance": {"const": cls.provenance},
                "class_name": {"const": cls.__name__},
                "parameters": {
                    "type": "object",
                    "properties": cls._param_schema(),
                    "required": list(cls.default_distribution()._param_dict())
                }
            },
            "required": ["implements", "provenance", "class_name", "parameters"]
        }

    @classmethod
    def from_dict(cls, dist_dict: dict) -> BaseDistribution:
        """Create a distribution from a dictionary."""
        return cls(**dist_dict["parameters"])

    def information_criterion(self, values: Iterable) -> float:  # pylint: disable=unused-argument
        """Get the AIC value for a particular set of values.

        Parameters
        ----------
        values: array_like
            Values to determine the AIC value of.
        """
        return 0.0

    @classmethod
    def matches_name(cls, name: str) -> bool:
        """Check whether the name matches the distribution.

        Parameters
        ----------
        name: str
            Name to match to the distribution.

        Returns
        -------
        bool:
            Whether the name matches.
        """
        assert cls.implements != "unknown", f"Internal error in class {cls.__name__}"
        return name in (cls.implements.split(".")[1],
                        cls.implements,
                        cls.__name__,
                        )

    @classmethod
    @abstractmethod
    def default_distribution(cls) -> BaseDistribution:
        """Get a distribution with default parameters."""
        return cls()


def metadist(
        implements: Optional[str] = None,
        provenance: Optional[str] = None,
        var_type: Optional[Union[str, list[str]]] = None,
        is_unique: Optional[bool] = None,
        version: Optional[str] = None,
        privacy: Optional[str] = None):
    """Decorate class to create a distribution with the right properties.

    Parameters
    ----------
    implements:
        The distribution ID that it implements, e.g. core.uniform, core.regex.
    provenance:
        Where the distribution came from, which package/plugin implemented it.
    var_type:
        Variable type of the distribution, e.g. continuous, categorical, string.
    is_unique:
        Whether the distribution is unique or not.
    privacy:
        Privacy class/implementation of the distribution.
    version:
        Version of the distribution. Increment this to ensure that compatibility is
        properly handled.

    Returns
    -------
    cls:
        Class with the appropriate class variables.
    """
    def _wrap(cls):
        if implements is not None:
            cls.implements = implements
        if provenance is not None:
            cls.provenance = provenance
        if var_type is not None:
            cls.var_type = var_type
        if is_unique is not None:
            cls.is_unique = is_unique
        if privacy is not None:
            cls.privacy = privacy
        if version is not None:
            cls.version = version
        return cls
    return _wrap


class ScipyDistribution(BaseDistribution):
    """Base class for numerical Scipy distributions.

    This base class makes it easy to implement new numerical
    distributions. One could also use this base class for non-scipy
    distributions, in which case the distribution class should implement
    logpdf, rvs and fit methods.
    """

    @property
    def n_par(self) -> int:
        """int: Number of parameters for distribution."""
        return len(self.par)

    def __getattr__(self, attr: str):
        """Get attribute for easy access to parameters.

        Parameters
        ----------
        attr:
            Attribute to retrieve. If the attribute is a parameter
            name, then retrieve that parameter, otherwise use the default
            implementation for getting the attribute.

        Returns
        -------
        object:
            Parameter or attribute.
        """
        if attr != "par" and attr in self.par:
            return self.par[attr]
        return object.__getattribute__(self, attr)

    @classmethod
    def _fit(cls, values):
        if len(values) == 0:
            return cls.default_distribution()
        param = cls.dist_class.fit(values)
        return cls(*param)

    def _param_dict(self):
        return self.par

    def draw(self):
        return self.dist.rvs()

    def information_criterion(self, values):
        vals = self._to_series(values)
        if len(vals) == 0:
            return 2 * self.n_par
        return self._information_criterion(vals)

    def _information_criterion(self, values):
        return 2 * self.n_par - 2 * np.sum(self.dist.logpdf(values))


@metadist(is_unique=True)
class UniqueDistributionMixin(BaseDistribution):
    """Mixin class to make unique version of base distribution."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_set: set = set()

    def draw_reset(self):
        self.key_set = set()

    def draw(self) -> object:
        n_retry = 0
        while n_retry < 1e5:
            new_val = super().draw()
            if new_val not in self.key_set:
                self.key_set.add(new_val)
                return new_val
            n_retry += 1
        raise ValueError(f"Failed to draw unique string after {n_retry} tries.")

    def information_criterion(self, values):
        return 9999999



"""Distributions for variables.

These distributions can be fit to datasets/series so that the synthesis is
somewhat realistic. The concept of distributions here is not only for
numerical data, but also for generating strings for example.
"""  # pylint: disable=invalid-name

from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.continuous import (ExponentialDistribution,
                                               LogNormalDistribution,
                                               NormalDistribution,
                                               TruncatedNormalDistribution,
                                               UniformDistribution)
from metasynth.distribution.datetime import (UniformDateDistribution,
                                             UniformDateTimeDistribution,
                                             UniformTimeDistribution)
from metasynth.distribution.discrete import (DiscreteUniformDistribution,
                                             PoissonDistribution,
                                             UniqueKeyDistribution)
from metasynth.distribution.faker import (FakerDistribution, UniqueFakerDistribution,
                                          FreeTextDistribution)
from metasynth.distribution.regex import (RegexDistribution,
                                          UniqueRegexDistribution)
from metasynth.distribution.na import NADistribution
__all__ = [
    "MultinoulliDistribution", "UniformDistribution", "NormalDistribution",
    "LogNormalDistribution", "TruncatedNormalDistribution", "ExponentialDistribution",
    "DiscreteUniformDistribution", "PoissonDistribution", "UniqueKeyDistribution",
    "UniformDateDistribution", "UniformDateTimeDistribution", "UniformTimeDistribution",
    "FakerDistribution", "UniqueFakerDistribution", "RegexDistribution", "UniqueRegexDistribution",
    "NADistribution", "FreeTextDistribution"

]



"""Module with discrete distributions."""

from typing import Set

import numpy as np
from scipy.stats import poisson, randint

from metasynth.distribution.base import (metadist,
                                         ScipyDistribution)


@metadist(implements="core.discrete_uniform", var_type="discrete")
class DiscreteUniformDistribution(ScipyDistribution):
    """Integer uniform distribution.

    It differs from the floating point uniform distribution by
    being a discrete distribution instead.

    Parameters
    ----------
    low: int
        Lower bound (inclusive) of the uniform distribution.
    high: int
        Upper bound (exclusive) of the uniform distribution.
    """

    dist_class = randint

    def __init__(self, low: int, high: int):
        self.par = {"low": low, "high": high}
        self.dist = self.dist_class(low=low, high=high)

    def _information_criterion(self, values):
        return 2*self.n_par - 2*np.sum(self.dist.logpmf(values))

    @classmethod
    def _fit(cls, values):
        param = {"low": values.min(), "high": values.max()+1}
        return cls(**param)

    @classmethod
    def default_distribution(cls):
        return cls(0, 10)

    @classmethod
    def _param_schema(cls):
        return {
            "low": {"type": "integer"},
            "high": {"type": "integer"},
        }


@metadist(implements="core.poisson", var_type="discrete")
class PoissonDistribution(ScipyDistribution):
    """Poisson distribution."""

    dist_class = poisson

    def __init__(self, mu: float):
        self.par = {"mu": mu}
        self.dist = self.dist_class(mu=mu)

    def _information_criterion(self, values):
        return 2*self.n_par - 2*np.sum(self.dist.logpmf(values))

    @classmethod
    def _fit(cls, values):
        return cls(values.mean())

    @classmethod
    def default_distribution(cls):
        return cls(0.5)

    @classmethod
    def _param_schema(cls):
        return {
            "mu": {"type": "number"},
        }


@metadist(implements="core.unique_key", var_type="discrete", is_unique=True)
class UniqueKeyDistribution(ScipyDistribution):
    """Integer distribution with unique keys.

    Discrete distribution that ensures the uniqueness of the drawn values.

    Parameters
    ----------
    low: int
        Minimum value for the keys.
    consecutive: int
        1 if keys are consecutive and increasing, 0 otherwise.
    """

    def __init__(self, low: int, consecutive: int):
        self.par = {"low": low, "consecutive": consecutive}
        self.last_key = low - 1
        self.key_set: Set[int] = set()

    @classmethod
    def _fit(cls, values):
        low = values.min()
        high = values.max() + 1
        if len(values) == high-low and np.all(values.to_numpy() == np.arange(low, high)):
            return cls(low, 1)
        return cls(low, 0)

    def draw_reset(self):
        self.last_key = self.low - 1
        self.key_set = set()

    def draw(self):
        if self.consecutive == 1:
            self.last_key += 1
            return self.last_key

        while True:
            random_number = np.random.randint(self.low, self.low+2*len(self.key_set)+2)
            if random_number not in self.key_set:
                self.key_set.add(random_number)
                return random_number

    def _information_criterion(self, values):
        if values.min() < self.low:
            return 3+999*len(values)

        # If the values are not unique the fit is extremely bad.
        if len(set(values)) != len(values):
            return 3+999*len(values)

        low = values.min()
        high = values.max()+1

        if self.consecutive == 1:
            # Check if the values are truly
            if len(values) == high-low and np.all(values.to_numpy() == np.arange(low, high)):
                return 3
            return 3+999*len(values)

        n_choice = high - low

        # Probabilities go up like 1/n, 1/(n-1), 1/(n-2), ..., 1/2, 1
        return 5 - 2*np.sum(np.log(1/np.arange(n_choice, n_choice-len(values), -1)))

    @classmethod
    def default_distribution(cls):
        return cls(0, 0)

    @classmethod
    def _param_schema(cls):
        return {
            "low": {"type": "integer"},
            "high": {"type": "integer"},
        }



"""Distribution for structured strings, using regexes."""
from __future__ import annotations

from typing import Union, Optional

from regexmodel import RegexModel, NotFittedError

from metasynth.distribution.base import metadist, BaseDistribution, UniqueDistributionMixin


@metadist(implements="core.regex", var_type="string", version="2.0")
class RegexDistribution(BaseDistribution):
    """Distribution for structured strings.

    Main implementation details in the regexmodel package:
    https://github.com/sodascience/regexmodel

    This distribution tries to create a regex that fits the variable. This
    regex also contains statistical information about the probabilities if the
    regex has multiple options (e.g. a|b). The regex is only a subset what is provided
    by the python re package. What is currently implemented:

    - Parentheses with multiple options and no modifiers, e.g. ([a]|[b]|[c])
    - Square brackets without negation, e.g. [abc]
    - Ranges [A-Z], [a-z], [0-9], but not subranges (e.g. [0-3])
    - Repetition quantifiers (curly brackets) with minimum and maximum [A-Z]{3,6}, but not [A-Z]{6}.

    When fitting the RegexDistribution using the fit method, pay attention to the
    count_thres and method parameters. By default these will be dynamic and take reasonable
    values for the input, but in some cases it can be important to set them manually.
    The count_thres parameter sets the minimum number of times a regex element needs to
    be used. So, if count_thres=2, and there is only one value starting with "a", then the regex
    will never start with "a". In effect, a higher value will provide more privacy, less
    utility and a faster fit. The other parameter "method" has a small effect on the accuracy
    of the regex, and a larger effect on the worst case time consumption for fitting.
    Set to "accurate" for the best result, and "fast" for the fastest result.

    Examples that this distribution should work reasonably for are: email,
    ID's, telephone numbers, ip addresses, etc.

    Parameters
    ----------
    regex_data:
        Valid inputs for the regex model are:
        - str: String with a regex (that falls within the specifications, see above).
        - dict: Serialized version of the regex model, as it is coming from a JSON file.
        - RegexModel: Initialized regex model.
    """

    def __init__(self, regex_data: Union[str, dict, RegexModel]):
        self.regex_model = RegexModel(regex_data)

    @classmethod
    def _fit(cls, values, count_thres: Optional[int] = None, method: str = "auto"):
        """Fit a regex to structured strings.

        Arguments
        ---------
        values:
            Values to be fitted (pl.Series).
        count_thres:
            Threshold for regex elements, so that a regex element can only be used if
            the number of values satisfying said element is higher than the threshold.
        method:
            Method for fitting the regex model. Possible values are ["accurate", "fast", "auto"]
            The "auto" method switches between the "accurate" and "fast" methods depending on
            the number of characters (fast if #char > 10000) in the series.
        """
        if method == "auto":
            if values.str.lengths().mean() > 10:
                method = "fast"
            else:
                method = "accurate"

        # Make count_thres ~= #values/100 up to 50 if in auto mode.
        if count_thres is None:
            count_thres = min(50, max(2, round(len(values)/50)))

        # Try to fit the values, if it cannot be fit, then use the default distribution.
        try:
            model = RegexModel.fit(values, count_thres=count_thres, method=method)
        except NotFittedError:
            return cls.default_distribution()
        return cls(model)

    def draw(self):
        return self.regex_model.draw()

    def _param_dict(self):
        return {"regex_data": self.regex_model.serialize()}

    @property
    def _params_formatted(self):
        return f"\t- regex: {self.regex_model.regex}"

    @classmethod
    def _param_schema(cls):
        return {
            "regex": {
                "type": "string"
            },
            "counts": {
                "type": "array"
            }
        }

    @classmethod
    def default_distribution(cls):
        return cls(r"[ABC][0-9]{3,4}")


@metadist(implements="core.unique_regex", var_type="string", is_unique=True)
class UniqueRegexDistribution(UniqueDistributionMixin, RegexDistribution):
    """Unique variant of the regex distribution."""



"""Module with privacy classes to be used for creating GMF files."""

from abc import abstractmethod, ABC
from typing import Type, Union

from metasynth.distribution.base import BaseDistribution


class BasePrivacy(ABC):
    """Base class for privacy level.

    Derived classes should at least set the class variable
    name and implement the to_dict method.
    """

    name = "unknown_privacy"

    def __init__(self):
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        """Create a dictionary that gives the privacy type, and parameters."""
        return {
            "type": self.name,
            "parameters": {},
        }

    def is_compatible(self, dist: Union[BaseDistribution, Type[BaseDistribution]]) -> bool:
        """Check whether the distribution has the same privacy.

        Arguments
        ---------
        dist:
            Distribution to check.
        """
        return dist.privacy == self.name

    @property
    def fit_kwargs(self):
        """Fitting arguments that need to be supplied to the distribution.

        For example epsilon in the case of differential privacy.
        """
        return {}


class BasicPrivacy(BasePrivacy):
    """No privacy class, which uses statistically optimal distributions."""

    name = "none"

    def to_dict(self) -> dict:
        return BasePrivacy.to_dict(self)



"""Variable module that creates metadata variables."""  # pylint: disable=invalid-name

from __future__ import annotations

from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
import polars as pl

from metasynth.distribution.base import BaseDistribution
from metasynth.privacy import BasePrivacy, BasicPrivacy
from metasynth.provider import (BaseDistributionProvider,
                                DistributionProviderList)


class MetaVar():
    """Meta data variable.

    Acts as a base class for specific types of variables, but also as a
    launching pad for detecting its type.

    Parameters
    ----------
    var_type:
        Variable type as a string, e.g. continuous, string, etc.
    series:
        Series to create the variable from. Series is None by default and in
        this case the value is ignored. If it is not supplied, then the
        variable cannot be fit.
    name:
        Name of the variable/column.
    distribution:
        Distribution to draw random values from. Can also be set by using the
        fit method.
    prop_missing:
        Proportion of the series that are missing/NA.
    dtype:
        Type of the original values, e.g. int64, float, etc. Used for type-casting
        back.
    description:
        User provided description of the variable.
    """

    dtype = "unknown"

    def __init__(self,  # pylint: disable=too-many-arguments
                 var_type: str,
                 series: Optional[Union[pl.Series, pd.Series]] = None,
                 name: Optional[str] = None,
                 distribution: Optional[BaseDistribution] = None,
                 prop_missing: Optional[float] = None,
                 dtype: Optional[str] = None,
                 description: Optional[str] = None):
        self.var_type = var_type
        self.prop_missing = prop_missing
        if series is None:
            self.name = name
            if dtype is not None:
                self.dtype = dtype
        else:
            series = _to_polars(series)
            self.name = series.name
            if prop_missing is None:
                self.prop_missing = (
                    len(series) - len(series.drop_nulls())) / len(series)
            self.dtype = str(series.dtype)

        self.series = series
        self.distribution = distribution
        self.description = description

        if self.prop_missing is None:
            raise ValueError(f"Error while initializing variable {self.name}."
                             " prop_missing is None.")
        if self.prop_missing < -1e-8 or self.prop_missing > 1+1e-8:
            raise ValueError(f"Cannot create variable '{self.name}' with proportion missing "
                             "outside range [0, 1]")

    @classmethod
    def detect(cls,
               series_or_dataframe: Union[pd.Series,
                                          pl.Series,
                                          pl.DataFrame],
               description: Optional[str] = None,
               prop_missing: Optional[float] = None):
        """Detect variable class(es) of series or dataframe.

        Parameters
        ----------
        series_or_dataframe: pd.Series or pd.Dataframe
            If the variable is a pandas Series, then find the correct
            variable type and create an instance of that variable.
            If a Dataframe is supplied instead, a list of of variables is
            returned: one for each column in the dataframe.
        description:
            User description of the variable.
        prop_missing:
            Proportion of the values missing. If None, detect it from the series.
            Otherwise prop_missing should be a float between 0 and 1.

        Returns
        -------
        MetaVar:
            It returns a meta data variable of the correct type.
        """
        if isinstance(series_or_dataframe, (pl.DataFrame, pd.DataFrame)):
            if isinstance(series_or_dataframe, pd.DataFrame):
                return [MetaVar.detect(series_or_dataframe[col])
                        for col in series_or_dataframe]
            return [MetaVar.detect(series) for series in series_or_dataframe]

        series = _to_polars(series_or_dataframe)

        try:
            var_type = pl.datatypes.dtype_to_py_type(series.dtype).__name__
        except NotImplementedError:
            var_type = pl.datatypes.dtype_to_ffiname(series.dtype)
        return cls(cls.get_var_type(var_type), series,
                   description=description, prop_missing=prop_missing)

    @staticmethod
    def get_var_type(polars_dtype: str) -> str:
        """Convert polars dtype to MetaSynth variable type."""
        convert_dict = {
            "int": "discrete",
            "float": "continuous",
            "date": "date",
            "datetime": "datetime",
            "time": "time",
            "str": "string",
            "categorical": "categorical"
        }
        try:
            return convert_dict[polars_dtype]
        except KeyError as exc:
            raise ValueError(
                f"Unsupported polars type '{polars_dtype}") from exc

    def to_dict(self) -> Dict[str, Any]:
        """Create a dictionary from the variable."""
        if self.distribution is None:
            dist_dict = {}
        else:
            dist_dict = self.distribution.to_dict()
        var_dict = {
            "name": self.name,
            "type": self.var_type,
            "dtype": self.dtype,
            "prop_missing": self.prop_missing,
            "distribution": dist_dict,
        }
        if self.description is not None:
            var_dict["description"] = self.description
        return var_dict

    def __str__(self) -> str:
        """Return an easy to read formatted string for the variable."""
        description = f'Description: "{self.description}"\n' if self.description else ""

        if self.distribution is None:
            distribution_formatted = "No distribution information available"
        else:
            distribution_formatted = "\n".join(
                "\t" + line for line in str(self.distribution).split("\n")
            )

        return (
            f'"{self.name}"\n'
            f'{description}'
            f'- Variable Type: {self.var_type}\n'
            f'- Data Type: {self.dtype}\n'
            f'- Proportion of Missing Values: {self.prop_missing:.4f}\n'
            f'- Distribution:\n{distribution_formatted}\n'
        )

    def fit(self,  # pylint: disable=too-many-arguments
            dist: Optional[Union[str, BaseDistribution, type]] = None,
            dist_providers: Union[str, type,
                                  BaseDistributionProvider] = "builtin",
            privacy: BasePrivacy = BasicPrivacy(),
            unique: Optional[bool] = None,
            fit_kwargs: Optional[dict] = None):
        """Fit distributions to the data.

        If multiple distributions are available for the current data type,
        use the one that fits the data the best.

        While it has no arguments or return values, it will set the
        distribution attribute to the most suitable distribution.

        Parameters
        ----------
        dist:
            The distribution to fit. In case of a string, search for it
            using the aliases of all distributions. Otherwise use the
            supplied distribution (class). Examples of allowed strings are:
            "normal", "uniform", "faker.city.nl_NL". If not supplied, fit
            the best available distribution for the variable type.
        dist_providers:
            Distribution providers that are used for fitting.
        privacy:
            Privacy level to use for fitting the series.
        unique:
            Whether the variable should be unique. If not supplied, it will be
            inferred from the data.
        fit_kwargs:
            Extra options for distributions during the fitting stage.
        """
        if self.series is None:
            raise ValueError("Cannot fit distribution if we don't have the"
                             "original data.")

        provider_list = DistributionProviderList(dist_providers)
        self.distribution = provider_list.fit(
            self.series, self.var_type, dist, privacy, unique, fit_kwargs)

    def draw(self) -> Any:
        """Draw a random item for the variable in whatever type is required."""
        if self.distribution is None:
            raise ValueError("Cannot draw without distribution")

        # Return NA's -> None
        if self.prop_missing is not None and np.random.rand() < self.prop_missing:
            return None
        return self.distribution.draw()

    def draw_series(self, n: int) -> pl.Series:
        """Draw a new synthetic series from the metadata.

        Parameters
        ----------
        n:
            Length of the series to be created.

        Returns
        -------
        pandas.Series:
            Pandas series with the synthetic data.
        """
        if not isinstance(self.distribution, BaseDistribution):
            raise ValueError("Cannot draw without distribution.")
        self.distribution.draw_reset()
        value_list = [self.draw() for _ in range(n)]
        if "Categorical" in self.dtype:
            return pl.Series(value_list, dtype=pl.Categorical)
        return pl.Series(value_list)

    @classmethod
    def from_dict(cls,
                  var_dict: Dict[str, Any],
                  distribution_providers: Union[
                      None, str, type[BaseDistributionProvider],
                      BaseDistributionProvider] = None) -> MetaVar:
        """Restore variable from dictionary.

        Parameters
        ----------
        distribution_providers:
            Distribution providers to use to create the variable. If None,
            use all installed/available distribution providers.
        var_dict:
            This dictionary contains all the variable and distribution
            information to recreate it from scratch.

        Returns
        -------
        MetaVar:
            Initialized metadata variable.
        """
        provider_list = DistributionProviderList(distribution_providers)
        dist = provider_list.from_dict(var_dict)
        return cls(
            var_dict["type"],
            name=var_dict["name"],
            distribution=dist,
            prop_missing=var_dict["prop_missing"], dtype=var_dict["dtype"],
            description=var_dict.get("description", None)
        )


def _to_polars(series: Union[pd.Series, pl.Series]) -> pl.Series:
    if isinstance(series, pl.Series):
        return series
    if len(series.dropna()) == 0:
        series = pl.Series(name=series.name,
                           values=[None for _ in range(len(series))])
    else:
        series = pl.Series(series)
    return series



"""Legacy distributions that will be removed in the future."""

from metasynth.distribution.legacy.regex import RegexDistribution
from metasynth.distribution.legacy.regex import UniqueRegexDistribution


__all__ = ["RegexDistribution", "UniqueRegexDistribution"]



"""Tools for validating distribution/GMF file output."""

from __future__ import annotations

from copy import deepcopy

try:
    from importlib_metadata import entry_points
except ImportError:
    from importlib.metadata import entry_points  # type: ignore

import jsonschema

from metasynth.provider import get_distribution_provider
from metasynth.distribution.na import NADistribution


SCHEMA_BASE = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "http://sodascience.github.io/generative_metadata_format/core/1.0.0/generative_metadata_format",  # noqa # pylint: disable=line-too-long
    "type": "object",
    "properties": {
        "n_rows": {"type": "number"},
        "n_columns": {"type": "number"},
        "provenance": {
            "type": "object",
            "properties": {
                "created by": {"type": "object"},
                "creation time": {"type": "string"}
            },
            "required": ["created by", "creation time"]
        },
        "vars": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "type": {"enum": ["discrete", "continuous", "string", "categorical", "date",
                                      "datetime", "time"]},
                    "dtype": {"type": "string"},
                    "prop_missing": {"type": "number"},
                    "distribution": {
                        "$ref": "#/$defs/all_dist_def"
                    }
                }
            },
            "required": ["name", "type", "dtype", "provenance", "prop_missing", "distribution"]
        }
    },
    "required": ["n_rows", "n_columns", "vars"],
}


def validate_gmf_dict(gmf_dict: dict):
    """Validate a JSON dictionary of a dataset as it would be written to a GMF file.

    Make sure that you have used the _jsonify function to convert numpy arrays to
    lists, etc.

    Arguments
    ---------
    gmf_dict:
        Dictionary containing the MetaSynth output for a dataset.
    """
    packages = [entry.name for entry in entry_points(group="metasynth.distribution_provider")]
    schema = create_schema(packages)
    jsonschema.validate(gmf_dict, schema)


def create_schema(packages: list[str]) -> dict:
    """Create JSON Schema to validate a GMF file.

    Arguments
    ---------
    packages:
        List of packages to create the schema with.

    Returns
    -------
    schema:
        Schema containing all the distributions in the distribution packages.
    """
    defs: list[dict] = []
    for package_name in packages:
        pkg = get_distribution_provider(package_name)
        for dist in pkg.distributions:
            defs.append(dist.schema())
    defs.append(NADistribution.schema())

    schema = deepcopy(SCHEMA_BASE)
    schema.update({"$defs": {"all_dist_def": {"anyOf": defs}}})
    return schema



"""Module containing the class with NA distributions."""

import polars as pl

from metasynth.distribution.base import BaseDistribution, metadist


@metadist(implements="core.na", var_type="string")
class NADistribution(BaseDistribution):
    """Distribution that will only ever give back NA."""

    @classmethod
    def _fit(cls, values: pl.Series) -> BaseDistribution:
        return cls()

    @classmethod
    def default_distribution(cls) -> BaseDistribution:
        return cls()

    def draw(self):
        return None

    def _param_dict(self):
        return {}

    @classmethod
    def _param_schema(cls):
        return {}



import datetime as dt

from pytest import mark
import pandas as pd
import polars as pl

from metasynth.distribution.datetime import UniformDateDistribution, UniformDateTimeDistribution
from metasynth.distribution.datetime import UniformTimeDistribution

all_precision = ["microseconds", "seconds", "minutes", "hours"]
start = ["10", ""]


@mark.parametrize(
    "start,end,precision", [
        ("10:23:45.293852", "16:23:45.293852", "microseconds"),
        ("10:23:45", "16:23:45", "seconds"),
        ("10:23:00", "16:23:00", "minutes"),
        ("10:00:00", "16:00:00", "hours"),
    ]
)
@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_time(start, end, precision, series_type):
    begin_iso, end_iso = dt.time.fromisoformat(start), dt.time.fromisoformat(end)
    dist = UniformTimeDistribution(start, end, precision)
    all_times = []
    for _ in range(100):
        new_time = dist.draw()
        assert isinstance(new_time, dt.time)
        assert new_time >= begin_iso
        assert new_time <= end_iso
        for prec in all_precision:
            if prec == precision:
                break
            assert getattr(new_time, prec[:-1]) == 0
        all_times.append(new_time)
    series = series_type(all_times)
    new_dist = UniformTimeDistribution.fit(series)
    assert new_dist.start >= dist.start
    assert new_dist.end <= dist.end
    assert new_dist.precision == dist.precision


@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_date(series_type):
    begin_date, end_date = "2020-07-09", "2023-08-19"
    begin_iso, end_iso = dt.date.fromisoformat(begin_date), dt.date.fromisoformat(end_date)

    dist = UniformDateDistribution(begin_date, end_date)
    all_dates = []
    for _ in range(100):
        new_date = dist.draw()
        assert isinstance(new_date, dt.date)
        assert new_date >= begin_iso
        assert new_date <= end_iso
        all_dates.append(new_date)
    series = series_type(all_dates)
    new_dist = UniformDateDistribution.fit(series)
    assert new_dist.start >= dist.start
    assert new_dist.end <= dist.end
    assert new_dist.precision == dist.precision


@mark.parametrize(
    "start,end,precision", [
        ("2020-07-09 10:23:45.293852", "2020-08-09 16:23:45.293852", "microseconds"),
        ("2020-07-09 10:23:45", "2020-08-09 16:23:45", "seconds"),
        ("2020-07-09 10:23:00", "2020-08-09 16:23:00", "minutes"),
        ("2020-07-09 10:00:00", "2020-08-09 16:00:00", "hours"),
    ]
)
@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_datetime(start, end, precision, series_type):
    begin_iso, end_iso = dt.datetime.fromisoformat(start), dt.datetime.fromisoformat(end)
    dist = UniformDateTimeDistribution(start, end, precision)
    all_datetimes = []
    for _ in range(100):
        new_time = dist.draw()
        assert isinstance(new_time, dt.datetime)
        assert new_time >= begin_iso
        assert new_time <= end_iso
        for prec in all_precision:
            if prec == precision:
                break
            assert getattr(new_time, prec[:-1]) == 0
        all_datetimes.append(new_time)

    series = series_type(all_datetimes)
    new_dist = UniformDateTimeDistribution.fit(series)
    assert new_dist.start >= dist.start
    assert new_dist.end <= dist.end
    assert new_dist.precision == dist.precision



"""Module containing string distributions."""

from __future__ import annotations

import random
import re
import string
from abc import ABC, abstractmethod
from typing import Dict, Iterable, Optional, Sequence, Tuple

import numpy as np
from sklearn.feature_extraction.text import CountVectorizer

from metasynth.distribution.legacy.regex.optimizer import RegexOptimizer


class BaseRegexElement(ABC):
    """Base class for regex elements."""

    frac_used = 1.0

    def __init__(self, frac_used: float):
        if frac_used > 1 or frac_used < 0:
            raise ValueError(f"Error initializing RegexElement with fraction used {frac_used}."
                             f"Set the fraction to be used between 0 (never) and 1 (always).")
        self.frac_used = frac_used

    @abstractmethod
    def information_budget(self, regex_stat: Dict) -> float:
        """Addition to the AIC for the current regex."""

    @property
    def n_param(self) -> int:
        """int: The number of 'parameters' of the regex element."""
        return 2

    @classmethod
    @abstractmethod
    def fit(cls, values: Sequence[str]) -> Tuple[
            Tuple[Iterable[str], Iterable[str]], float, BaseRegexElement]:
        """Match the regex to the values, using as many positions as possible.

        Parameters
        ----------
        values: array_like of str
            Remaining values to match the regex to.

        Returns
        -------
        BaseRegexElement:
            The optimized regex.
        """

    @classmethod
    @abstractmethod
    def fit_start(cls, values: Sequence[str]) -> Tuple[Iterable[str], BaseRegexElement]:
        """Match the regex to the values, sequential greedy algotithm from the left.

        Parameters
        ----------
        values: array_like of str
            Remaining values to match the regex to.

        Returns
        -------
        BaseRegexElement:
            The optimized regex.
        """

    @abstractmethod
    def _draw(self) -> str:
        """Draw a random string from the regex element.

        Returns
        -------
        str:
            The string drawn from the regex distribution element.
        """

    def draw(self) -> str:
        """Draw a random string, considering how often the element is used.

        Returns
        -------
        random_string: String randomly drawn from the regex.
        """
        if np.random.rand() < self.frac_used:
            return self._draw()
        return ""

    @classmethod
    @abstractmethod
    def from_string(cls, regex_str: str, frac_used: float = 1.0
                    ) -> Optional[Tuple[BaseRegexElement, str]]:
        """Create a regex object from a regex string.

        Parameters
        ----------
        regex_str: str
            Regex string to process. Only single elements are supported.
        frac_used: float
            Fraction of the generated values for which the regex element is used.

        Returns
        -------
        Tuple[BaseRegexElement, str] or None:
            If the regex element is compatible with the string, then return the regex element
            that corresponds to the string. Otherwise return None.
        """


class BaseRegexClass(BaseRegexElement):
    """Base class for repeating regex elements.

    Repeating elements are e.g. digits, letter, alphanumeric characters, etc.

    Parameters
    ----------
    min_digit: int
        Minimum number of repeats of the element.
    max_digit: int
        Maximum number of repeats of the element.
    frac_used: Fraction of values that are matched at all.
    """

    regex_str = r""
    match_str = r""
    base_regex = r""

    def __init__(self, min_digit: int, max_digit: int, frac_used: float = 1.0):
        super().__init__(frac_used)
        self.min_digit = min_digit
        self.max_digit = max_digit
        self._prepare_regex()

    def _prepare_regex(self):
        regex_str = self.base_regex + r"{" + str(max(1, self.min_digit))
        regex_str += r"," + str(max(1, self.max_digit)) + r"}"
        self.match_regex = re.compile(regex_str)

    @property
    @abstractmethod
    def digit_options(self) -> int:
        """Return the number of options per digit."""
        return 0

    @classmethod
    def fit_start(cls, values: Sequence[str]) -> Tuple[Iterable[str], BaseRegexElement]:
        regex = re.compile(cls.regex_str)
        new_values = []

        min_match = 99999
        max_match = 0
        n_vals_matched = 0
        for val in values:
            match = regex.search(val)
            if match is None:
                n_match = 0
            else:
                n_match = match.span()[1]
                n_vals_matched += 1
                min_match = min(n_match, min_match)
                max_match = max(n_match, max_match)
            new_values.append(val[n_match:])
        if min_match == 99999:
            min_match = 0
            max_match = 1
        return new_values, cls(min_match, max_match, n_vals_matched/len(values))

    @classmethod
    def fit(cls, values: Sequence[str]) -> Tuple[
            Tuple[Iterable[str], Iterable[str]], float, BaseRegexElement]:
        spans = cls.all_spans(values)
        optimizer = RegexOptimizer(values, spans)
        optimizer.optimize()
        stats = optimizer.statistics
        new_regex = cls(stats["min_len"], stats["max_len"], frac_used=stats["frac_used"])
        delta_energy = RegexOptimizer.energy_from_values(values) - optimizer.energy
        info_budget = new_regex.information_budget(stats)
        return optimizer.new_values, delta_energy/info_budget, new_regex

    # def check_split(self, stats):
    #     digit_counts = stats["digits_used"]
    #     for digit_split in range(self.min_digit, self.max_digit):
    #         digits_used_left = np.copy(digit_counts)
    #         digits_used_left[digit_split] += np.sum(digit_counts[digit_split+1:])
    #         digits_used_left[digit_split+1:] = 0
    #         stat_left = {
    #             "n_values": stats["n_values"],
    #             "frac_used": self.frac_used,
    #             "digits_used": digits_used_left,
    #         }
    #         left_regex = self.__class__(self.min_digit, digit_split)
    #
    #         digits_used_right = np.copy(digit_counts)
    #         digits_used_right[0] = np.sum(digit_counts[:digit_split+1])
    #         digits_used_right[1:]
    #         right_regex = self.__class__(1, self.max_digit-digit_split+1)

    @classmethod
    def from_string(cls, regex_str: str, frac_used: float = 1.0):
        match = re.search(cls.match_str, regex_str)
        if match is None:
            return None
        if all(x is None for x in match.groups()):
            return (cls(1, 1, frac_used=frac_used), regex_str[match.span()[1]:])
        return (cls(int(match.groups()[0]), int(match.groups()[1]), frac_used=frac_used),
                regex_str[match.span()[1]:])

    @classmethod
    def all_spans(cls, values: Sequence[str]) -> Sequence[Sequence[Tuple[int, int]]]:
        """Return match spans for a sequence of strings.

        Parameters
        ----------
        values:
            String values to match the regex to.

        Returns
        -------
        List[List[Tuple[int, int]]]:
            A list of spans for each of the values.
        """
        match_regex = re.compile(cls.base_regex+r"+")
        return [
            [match.span() for match in match_regex.finditer(val)]
            for val in values
        ]

    def information_budget(self, regex_stat: dict) -> float:
        if self.frac_used < 1e-8:
            return 2*self.n_param
        budget = 2*self.n_param
        n_values = regex_stat["n_values"]
        digit_range = self.max_digit-self.min_digit+1
        p_choose = regex_stat["frac_used"]/digit_range
        digits_factor = np.log(p_choose)
        for i_digit in range(self.min_digit, self.max_digit+1):
            i_count = regex_stat["digits_used"][i_digit]
            options_factor = -i_digit*np.log(self.digit_options)
            budget -= 2*i_count*(options_factor + digits_factor)
        num_unused = regex_stat["digits_used"][0]
        if num_unused > 0:
            budget -= 2*num_unused * np.log(num_unused/n_values)
        return budget


class DigitRegex(BaseRegexClass):
    """Regex element for repeating digits."""

    regex_str = r"^\d{1,}"
    match_str = r"^\\d(?:{(\d*),(\d*)})?"
    base_regex = r"\d"

    def _draw(self):
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([str(np.random.randint(0, 10)) for _ in range(n_digit)])

    def __str__(self):
        return r"\d{"+str(self.min_digit)+","+str(self.max_digit)+"}"

    @property
    def digit_options(self) -> int:
        return 10


class AlphaNumericRegex(BaseRegexClass):
    """Regex element for repeating alphanumeric character."""

    regex_str = r"^\w{1,}"
    match_str = r"^\\w(?:{(\d*),(\d*)})?"
    base_regex = r"\w"

    def _draw(self):
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([random.choice(string.ascii_letters+string.digits) for _ in range(n_digit)])

    def __str__(self):
        return r"\w{"+str(self.min_digit)+","+str(self.max_digit)+"}"

    @property
    def digit_options(self) -> int:
        return len(string.ascii_letters+string.digits)


class LettersRegex(BaseRegexClass):
    """Regex element for repeating letters."""

    regex_str = r"^[a-zA-Z]{1,}"
    match_str = r"^\[a-zA-Z\](?:{(\d*),(\d*)})?"
    base_regex = r"[a-zA-Z]"

    def _draw(self):
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([random.choice(string.ascii_letters) for _ in range(n_digit)])

    def __str__(self):
        return "[a-zA-Z]{"+str(self.min_digit)+","+str(self.max_digit)+"}"

    @property
    def digit_options(self) -> int:
        return len(string.ascii_letters)


class LowercaseRegex(BaseRegexClass):
    """Regex element for repeating lowercase letters."""

    regex_str = r"^[a-z]{1,}"
    match_str = r"^\[a-z\](?:{(\d*),(\d*)})?"
    base_regex = r"[a-z]"

    def _draw(self):
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([random.choice(string.ascii_lowercase) for _ in range(n_digit)])

    def __str__(self):
        return "[a-z]{"+str(self.min_digit)+","+str(self.max_digit)+"}"

    @property
    def digit_options(self) -> int:
        return len(string.ascii_lowercase)


class UppercaseRegex(BaseRegexClass):
    """Regex element for repeating uppercase letters."""

    regex_str = r"^[A-Z]{1,}"
    match_str = r"^\[A-Z\](?:{(\d*),(\d*)})?"
    base_regex = r"[A-Z]"

    def _draw(self):
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([random.choice(string.ascii_uppercase) for _ in range(n_digit)])

    def __str__(self):
        return "[A-Z]{"+str(self.min_digit)+","+str(self.max_digit)+"}"

    @property
    def digit_options(self) -> int:
        return len(string.ascii_uppercase)


class AnyRegex(BaseRegexClass):
    """Regex that matches any character.

    For generation, it uses string.printable characters plus any additional
    ones that are manually set.

    Parameters
    ----------
    min_digit:
        Minimum number of digits for the regex.
    max_digit:
        Maximum number of digits for the regex.
    extra_char:
        Extra characters to draw from outside of string.printable.
    frac_used:
        Fraction of the values containing the regex.
    """

    def __init__(self, min_digit: int, max_digit: int,  # pylint: disable=super-init-not-called
                 extra_char: Optional[set[str]] = None,
                 frac_used: float = 1.0):
        self.extra_char = set() if extra_char is None else extra_char
        super().__init__(min_digit, max_digit, frac_used)

    def _prepare_regex(self):
        self.all_char = string.printable + "".join(self.extra_char)
        self.all_char_set = set(self.all_char)

    @classmethod
    def fit_start(cls, values: Sequence[str]) -> Tuple[Iterable[str], BaseRegexElement]:
        new_values, _, regex = cls.fit(values)
        return new_values[1], regex

    @classmethod
    def all_spans(cls, values: Sequence[str]) -> Sequence[Sequence[Tuple[int, int]]]:
        return [[(0, len(v))] if len(v) > 0 else [] for v in values]

    @property
    def n_param(self) -> int:
        return 2 + len(self.extra_char)

    @property
    def digit_options(self) -> int:
        return len(string.printable) + len(self.extra_char)

    def _draw(self) -> str:
        n_digit = np.random.randint(self.min_digit, self.max_digit+1)
        return "".join([random.choice(self.all_char) for _ in range(n_digit)])

    @classmethod
    def from_string(cls, regex_str: str, frac_used: float = 1.0):
        match = re.search(r"^\.\[(.*)\](?:\{(\d+),(\d+)\})?", regex_str)
        if match is None:
            return None
        groups = match.groups()
        extra_char = set(groups[0])

        if all(x is None for x in groups[1:]):
            return cls(1, 1, extra_char), regex_str[match.span()[1]:]
        min_digit = int(groups[1])
        max_digit = int(groups[2])
        return (cls(min_digit, max_digit, extra_char, frac_used=frac_used),
                regex_str[match.span()[1]:])

    def __str__(self):
        return f".[{''.join(self.extra_char)}]{{{self.min_digit},{self.max_digit}}}"


class SingleRegex(BaseRegexElement):
    """Regex element that is a choice on a set of (any) characters.

    Parameters
    ----------
    character_selection:
        The selection of characters available. An empty character ''
        is also allowed.
    """

    def __init__(self, character_selection: list[str], frac_used: float = 1.0):
        super().__init__(frac_used)
        self.character_selection = list(sorted(character_selection))

    def _draw(self) -> str:
        return np.random.choice(self.character_selection)

    @property
    def n_param(self) -> int:
        return 1 + len(self.character_selection)

    def information_budget(self, regex_stat) -> float:
        if self.frac_used < 1e-8:
            return 2*self.n_param
        n_options = len(self.character_selection)
        n_unused = regex_stat["digits_used"][0]
        n_used = regex_stat["digits_used"][1]

        budget_used = n_used*np.log(self.frac_used/n_options)
        if 1-self.frac_used > 1e-8:
            budget_unused = n_unused*np.log(1-self.frac_used)
        else:
            budget_unused = 0
        return 2*self.n_param - 2*(budget_used+budget_unused)

    @classmethod
    def fit_start(cls, values: Sequence[str]) -> Tuple[Iterable[str], SingleRegex]:
        first_chars = [v[0] for v in values if len(v) > 0]
        new_values = [v[1:] for v in values]
        return new_values, cls(list(set(first_chars)), len(first_chars)/len(values))

    @classmethod
    def fit(cls, values):
        vec = CountVectorizer(analyzer="char", lowercase=False)
        counts = vec.fit_transform(values).getnnz(0)
        count_dict = {c: counts[i] for c, i in vec.vocabulary_.items()}
        sorted_counts = sorted(count_dict.items(), key=lambda x: x[1], reverse=True)

        cur_characters = []
        best_solution = (None, -10.0)
        while True:
            new_solution = _find_best_gradient(values, sorted_counts, cur_characters)

            # Check to see if no character is left.
            if new_solution[0] is None:
                break

            new_char, new_grad, optimizer = new_solution
            if new_grad > best_solution[1]:
                best_solution = (optimizer, new_grad)
                cur_characters.append(new_char)
            else:
                break
        optimizer, gradient = best_solution
        return optimizer.new_values, gradient, cls(cur_characters,
                                                   optimizer.statistics["frac_used"])

    def __str__(self):
        return "["+"".join(self.character_selection)+"]"

    @classmethod
    def from_string(cls, regex_str, frac_used=1.0):
        match = re.search(r"^\[(.+)\]", regex_str)
        if match is None:
            return None
        return cls(list(match.groups()[0]), frac_used), regex_str[match.span()[1]:]


def _create_spans_char(values, *characters):
    """Create the spans for the values given a number of characters for the match."""
    char_list = "".join(re.escape(c) for c in characters)
    regex = re.compile(r"[" + char_list + r"]")
    return [
        [match.span() for match in regex.finditer(val)]
        for val in values
    ]


def _find_best_gradient(values, sorted_counts, start_char=None):
    """Find the best gradient by adding one additional character to the set."""
    if start_char is None:
        start_char = []
    start_energy = RegexOptimizer.energy_from_values(values)
    best_solution = (None, -10.0, None)
    for character, _char_count in sorted_counts:
        if character in start_char:
            continue
        spans = _create_spans_char(values, character, *start_char)
        optimizer = RegexOptimizer(values, spans)
        optimizer.optimize()
        stats = optimizer.statistics
        regex = SingleRegex(start_char + [character], stats["frac_used"])
        info_budget = regex.information_budget(stats)
        energy_grad = (start_energy - optimizer.energy)/info_budget
        if energy_grad > best_solution[1]:
            best_solution = (character, energy_grad, optimizer)
    return best_solution



"""Distributions for date and time types."""

import datetime as dt
from abc import abstractmethod
from random import random
from typing import Any, Dict

import numpy as np

from metasynth.distribution.base import metadist, ScipyDistribution


def convert_numpy_datetime(time_obj: np.datetime64) -> dt.datetime:
    """Convert numpy datetime to python stdlib datetime.

    Parameters
    ----------
    time_obj:
        datetime to be converted.

    Returns
    -------
    datetime.datetime:
        Converted datetime.
    """
    unix_epoch = np.datetime64(0, 's')
    one_second = np.timedelta64(1, 's')
    seconds_since_epoch = (time_obj - unix_epoch) / one_second
    return dt.datetime.utcfromtimestamp(float(seconds_since_epoch))


class BaseUniformDistribution(ScipyDistribution):
    """Base class for all time/date/datetime uniform distributions."""

    precision_possibilities = ["microseconds", "seconds", "minutes", "hours", "days"]

    def __init__(self, start: Any, end: Any, precision: str = "microseconds"):
        if isinstance(start, str):
            start = self.fromisoformat(start)
        elif isinstance(start, np.datetime64):
            start = convert_numpy_datetime(start)
        if isinstance(end, str):
            end = self.fromisoformat(end)
        elif isinstance(end, np.datetime64):
            end = convert_numpy_datetime(end)
        self.par = {
            "start": start,
            "end": end,
            "precision": precision,
        }
        self.start = self.round(start)
        self.end = self.round(end)

    @classmethod
    def _fit(cls, values):
        return cls(values.min(), values.max(), cls._get_precision(values))

    @classmethod
    def _get_precision(cls, values):
        cur_precision = 0
        for precision in cls.precision_possibilities[:-1]:
            if not np.all([getattr(d, precision[:-1]) == 0 for d in values]):
                break
            cur_precision += 1
        return cls.precision_possibilities[cur_precision]

    def round(self, time_obj: Any) -> Any:
        """Round down any time object with the precision.

        Parameters
        ----------
        time_obj:
            Object to round down.

        Return
        ------
        obj:
            Time/date/datetime object rounded down to the measured precision.
        """
        for prec in self.precision_possibilities:
            if prec == self.precision:
                break
            time_obj = time_obj.replace(**{prec[:-1]: 0})
        try:
            time_obj = time_obj.replace(nanosecond=0)
        except TypeError:
            pass
        return time_obj

    def draw(self) -> dt.datetime:
        delta = self.end-self.start + self.minimum_delta
        new_time = random()*delta + self.start
        return self.round(new_time)

    @abstractmethod
    def fromisoformat(self, dt_obj: str):
        """Convert string to iso format."""

    @property
    def minimum_delta(self) -> dt.timedelta:
        """Get the minimum time delta."""
        return dt.timedelta(**{self.precision: 1})

    def information_criterion(self, values):
        return 0.0

    def _param_dict(self):
        return {
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "precision": self.precision,
        }


@metadist(implements="core.uniform_datetime", var_type="datetime")
class UniformDateTimeDistribution(BaseUniformDistribution):
    """Uniform DateTime distribution."""

    def fromisoformat(self, dt_obj: str) -> dt.datetime:
        return dt.datetime.fromisoformat(dt_obj)

    @classmethod
    def default_distribution(cls):
        return cls("2022-07-15T10:39:36", "2022-08-15T10:39:36", precision="seconds")

    @classmethod
    def _param_schema(cls):
        return {
            "start": {"type": "string"},
            "end": {"type": "string"},
            "precision": {"type": "string"},
        }


@metadist(implements="core.uniform_time", var_type="time")
class UniformTimeDistribution(BaseUniformDistribution):
    """Uniform time distribution."""

    def fromisoformat(self, dt_obj: str) -> dt.time:
        return dt.time.fromisoformat(dt_obj)

    @classmethod
    def default_distribution(cls):
        return cls("10:39:36", "18:39:36", precision="seconds")

    def draw(self):
        dt_begin = dt.datetime.combine(dt.datetime.today(), self.start)
        dt_end = dt.datetime.combine(dt.datetime.today(), self.end)
        delta = dt_end-dt_begin + self.minimum_delta
        return self.round((random()*delta + dt_begin).time())

    @classmethod
    def _param_schema(cls):
        return {
            "start": {"type": "string"},
            "end": {"type": "string"},
            "precision": {"type": "string"},
        }


@metadist(implements="core.uniform_date", var_type="date")
class UniformDateDistribution(BaseUniformDistribution):
    """Uniform date distribution."""

    precision_possibilities = ["days"]

    def __init__(self, start: Any, end: Any):
        super().__init__(start, end, precision="days")

    def fromisoformat(self, dt_obj: str) -> dt.date:
        return dt.date.fromisoformat(dt_obj)

    def round(self, time_obj):
        return time_obj

    def _param_dict(self) -> Dict:
        date_dict = BaseUniformDistribution._param_dict(self)
        del date_dict["precision"]
        return date_dict

    @classmethod
    def default_distribution(cls):
        return cls("1903-07-15", "1940-07-16")

    @classmethod
    def _fit(cls, values):
        return cls(values.min(), values.max())

    @classmethod
    def _param_schema(cls):
        return {
            "start": {"type": "string"},
            "end": {"type": "string"},
        }



"""CLI for generating synthetic data frames from a metasynth .json file."""
import argparse
import json
import pathlib
import pickle
import sys

try:  # Python < 3.10 (backport)
    from importlib_metadata import entry_points, version
except ImportError:
    from importlib.metadata import entry_points, version  # type: ignore [assignment]

from metasynth import MetaFrame
from metasynth.validation import create_schema

MAIN_HELP_MESSAGE = f"""
Metasynth CLI version {version("metasynth")}

Usage: metasynth [subcommand] [options]

Available subcommands:
    synthesize - generate synthetic data from a .json file
    jsonschema - generate json schema from distribution providers

Program information:
    -v, --version - display CLI version and exit
    -h, --help    - display this help file and exit
"""

ENTRYPOINTS = ["synthesize", "schema"]


def main() -> None:
    """CLI pointing to different entrypoints."""
    # show help by default, else consume first argument
    subcommand = "--help" if len(sys.argv) < 2 else sys.argv.pop(1)

    if subcommand in ["-h", "--help"]:
        print(MAIN_HELP_MESSAGE)
    elif subcommand in ["-v", "--version"]:
        print(f"Metasynth CLI version {version('metasynth')}")

    # find the subcommand in this module and run it!
    elif subcommand == "synthesize":
        synthesize()
    elif subcommand == "schema":
        schema()

    else:
        print(f"Invalid subcommand ({subcommand}). For help see metasynth --help")
        sys.exit(1)


def synthesize() -> None:
    """Program to generate synthetic data."""
    parser = argparse.ArgumentParser(
        prog="metasynth synthesize",
        description="Synthesize data from Generative Metadata Format .json file.",
    )
    parser.add_argument(
        "input",
        help="input file; typically .json adhering to the Generative Metadata Format",
        type=pathlib.Path,
    )
    parser.add_argument(
        "output",
        help="output file (.csv, .feather, .parquet, .pkl, or .xlsx)",
        nargs="?",
        type=pathlib.Path,
    )
    parser.add_argument(
        "-n", "--num_rows",
        help="number of rows to synthesize",
        default=None,
        type=int,
        required=False,
    )
    parser.add_argument(
        "-p", "--preview",
        help="preview six-row synthesized data frame in console and exit",
        action="store_true",
    )

    # parse the args without the subcommand
    args, _ = parser.parse_known_args()

    if not args.preview and not args.output:
        parser.error("Output file is required.")

    # Create the metaframe from the json file
    meta_frame = MetaFrame.from_json(args.input)

    if args.preview:
        # only print six rows and exit
        print(meta_frame.synthesize(6))
        return

    # Generate a data frame
    data_frame = meta_frame.synthesize(args.num_rows)

    # Store the dataframe to file
    if args.output.suffix == ".csv":
        data_frame.write_csv(args.output)
    elif args.output.suffix == ".feather":
        data_frame.write_ipc(args.output)
    elif args.output.suffix == ".parquet":
        data_frame.write_parquet(args.output)
    elif args.output.suffix == ".xlsx":
        data_frame.write_excel(args.output)
    elif args.output.suffix == ".pkl":
        with args.output.open("wb") as pkl_file:
            pickle.dump(data_frame, file=pkl_file)
    else:
        parser.error(
            f"Unsupported output file format ({args.output.suffix})."
            "Use .csv, .feather, .parquet, .pkl, or .xlsx.",
        )


def schema() -> None:
    """Program to generate json schema from dist providers."""
    parser = argparse.ArgumentParser(
        prog="metasynth schema",
        description="Create Generative Metadata Format schema and print to console.",
    )

    parser.add_argument(
        "plugins",
        help="Plugins to include in the generated schema (default builtin)",
        nargs="*"
    )

    parser.add_argument(
        "-l", "--list",
        help="display available plugins and quit",
        action="store_true",
    )

    # parse the args without the subcommand
    args, _ = parser.parse_known_args()

    # deduplicated list of plugins for schema
    plugins_avail = {entry.name for entry in entry_points(group="metasynth.distribution_provider")}

    if args.list:
        for a in plugins_avail:
            if a != "builtin":
                print(a)
        return

    plugins = {"builtin", *args.plugins}
    if len(plugins - plugins_avail) > 0:
        notfound = ", ".join(plugins - plugins_avail)
        pl_avail = ", ".join(plugins_avail - {"builtin"})
        errmsg = (
            f"\n  Requested plugin(s) not found: {notfound}"
            f"\n  Available plugins: {pl_avail}"
        )
        parser.error(errmsg)
    jsonschema = create_schema(list(plugins))
    print(json.dumps(jsonschema, indent=2))


if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
from metasynth.distribution.datetime import UniformDateTimeDistribution, UniformTimeDistribution
from metasynth.distribution.datetime import UniformDateDistribution
import wget
from pathlib import Path
import random


def get_demonstration_fp():
    demonstration_fp = Path("demonstration.csv")
    titanic_fp = Path("titanic.csv")
    if demonstration_fp.is_file():
        return demonstration_fp
    if not titanic_fp.is_file():
        wget.download("https://raw.githubusercontent.com/pandas-dev/pandas/main/doc/data/titanic.csv")
    df = pd.read_csv(titanic_fp)
    np.random.seed(1283742)
    random.seed(1928374)

    # Convert Age to a nullable integer.
    df["Age"] = df["Age"].round().astype("Int64")

    # Add a date column.
    date_dist = UniformDateDistribution.default_distribution()
    df["Birthday"] = [date_dist.draw() if np.random.rand() < 0.9 else pd.NA for _ in range(len(df))]

    # Add a time column.

    time_dist = UniformTimeDistribution.default_distribution()
    df["Board time"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA for _ in range(len(df))]

    # Add a datetime column
    time_dist = UniformDateTimeDistribution.default_distribution()
    df["Married since"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA for _ in range(len(df))]

    df["all_NA"] = [pd.NA for _ in range(len(df))]
    # Remove some columns for brevity and write to a file.
    df = df.drop(["SibSp", "Pclass", "Ticket", "Survived"], axis=1)
    df.to_csv(demonstration_fp, index=False)
    return demonstration_fp



"""Module containing the optimizer class for the regex distribution."""

from typing import Any, Dict, List, Sequence, Tuple

import numpy as np
import numpy.typing as npt

# Used for side assignment of values with no regex match
SIDE_LEFT = -1
SIDE_RIGHT = -2


class RegexOptimizer():
    r"""Optimizer class for fitting regexes to lists of strings.

    As input it takes the list of values and a list of matching spans
    for each of these values. With that it tries to find the optimal span assignment.
    It does so in a greedy way, where it optimizes each value step by step.

    As a result of the assignment, the values are split: one part left of the regex and
    another part right of the regex. The goal is to optimize the parts on the left and right
    to ensure that is as close to optimal as possible.

    The objective value of the optimization for this is defined for a list of values as follows:
    take the cumulative distribution of the lengths of the values > 0 and compute:
    \sum_i=1 \log (C_i + 1).
    , where C_i is the number of values with length i or more.

    Parameters
    ----------
    values:
        Values to which the regex is being applied.
    spans:
        Positions/spans where the regex can be applied.
    """

    def __init__(self, values: Sequence[str], spans: Sequence[Sequence[Tuple[int, int]]]):
        self.values = values
        self.spans = spans
        self.cur_solution = np.zeros(len(values), dtype=int)
        self.max_len = np.max([len(v) for v in values])
        self.left_cum_dist = np.zeros(self.max_len, dtype=int)
        self.right_cum_dist = np.zeros(self.max_len, dtype=int)
        for i_val, cur_val in enumerate(values):
            cur_span_choice = spans[i_val]
            if len(cur_span_choice) > 0:
                i_choice: int = int(np.argmax([x[1]-x[0] for x in cur_span_choice]))
                cur_span = cur_span_choice[i_choice]
                self.cur_solution[i_val] = i_choice
                self.left_cum_dist[:cur_span[0]] += 1
                self.right_cum_dist[:len(cur_val)-cur_span[1]] += 1
            else:
                self.cur_solution[i_val] = np.random.choice([SIDE_LEFT, SIDE_RIGHT])
                if self.cur_solution[i_val] == SIDE_LEFT:
                    self.left_cum_dist[:len(cur_val)] += 1
                else:
                    self.right_cum_dist[:len(cur_val)] += 1

    def optimize(self) -> None:
        """Optimize the positions until no improvement can be found."""
        while self.optimize_round():
            pass

    def optimize_round(self) -> bool:
        """Do one round of optimization.

        Returns
        -------
        has_changed: Whether the optimization round has changed any assignment.
        """
        has_changed: bool = False

        # Iterate over all values
        for i_val, cur_val in enumerate(self.values):
            cur_span_choice = self.spans[i_val]
            if len(cur_span_choice) == 0:
                chosen_side = self.cur_solution[i_val]

                # If assigned to the left, attempt right assignment.
                if chosen_side == SIDE_LEFT:
                    delta_energy = self.energy_move(self.left_cum_dist, len(cur_val), 0)
                    delta_energy += self.energy_move(self.right_cum_dist, 0, len(cur_val))
                    if delta_energy < -1e-8:
                        self.left_cum_dist[:len(cur_val)] -= 1
                        self.right_cum_dist[:len(cur_val)] += 1
                        self.cur_solution[i_val] = SIDE_RIGHT
                        has_changed = True

                # If assigned to the right, attempt left assignment
                elif chosen_side == SIDE_RIGHT:
                    delta_energy = self.energy_move(self.right_cum_dist, len(cur_val), 0)
                    delta_energy += self.energy_move(self.left_cum_dist, 0, len(cur_val))
                    if delta_energy < -1e-8:
                        self.left_cum_dist[:len(cur_val)] += 1
                        self.right_cum_dist[:len(cur_val)] -= 1
                        self.cur_solution[i_val] = SIDE_LEFT
                        has_changed = True
                continue

            # Only reached if there is at least one match.
            chosen_span = self.cur_solution[i_val]
            i_start, i_temp = cur_span_choice[chosen_span]
            i_end = len(cur_val)-i_temp
            best_choice = (chosen_span, 0.0)

            # Try all spans except the one we already have and find the one with the lowest energy.
            for j_choice, j_span in enumerate(cur_span_choice):
                if j_choice == self.cur_solution[i_val]:
                    continue
                j_start, j_temp = j_span
                j_end = len(cur_val)-j_temp
                delta_energy = (self.energy_move(self.left_cum_dist, i_start, j_start) +
                                self.energy_move(self.right_cum_dist, i_end, j_end))
                if delta_energy - best_choice[1] > -1e-8:
                    continue
                best_choice = (j_choice,  delta_energy)
            if best_choice[1] < 0:
                has_changed = True
                self.set_span(i_val, best_choice[0])
        return has_changed

    @property
    def energy(self) -> float:
        """Compute the energy of the current assignment/solution."""
        left_energy = np.sum(np.log(self.left_cum_dist+1))
        right_energy = np.sum(np.log(self.right_cum_dist+1))
        return left_energy + right_energy

    @staticmethod
    def energy_move(dist: npt.NDArray[np.int_], len_src: int, len_dst: int) -> float:
        """Compute the energy to change the length on one side.

        Parameters
        ----------
        dist:
            Cumulative distribution of the current lengths.
            (self.left_cum_dist or self.right_cum_dist)
        len_src:
            Current length on this side.
        len_dst:
            New length on the same side.

        Returns
        -------
        delta_energy:
            Difference in energy [after-before].
        """
        if len_dst > len_src:
            return np.sum(np.log((dist[len_src:len_dst]+2) /
                                 (dist[len_src:len_dst]+1)))
        if len_dst < len_src:
            assert np.all(dist[len_dst:len_src] > 0), (dist[len_dst:len_src], len_src, len_dst)
            return np.sum(np.log((dist[len_dst:len_src] /
                                  (dist[len_dst:len_src]+1))))
        return 0

    def _check_integrity(self):
        """Check the integrity of the assignments."""
        n_sum_left = 0
        n_sum_right = 0
        for i_val, cur_val in enumerate(self.values):
            i_span = self.cur_solution[i_val]
            if i_span == SIDE_LEFT:
                n_sum_left += len(cur_val)
            elif i_span == SIDE_RIGHT:
                n_sum_right += len(cur_val)
            else:
                n_sum_left += self.spans[i_val][i_span][0]
                n_sum_right += len(cur_val) - self.spans[i_val][i_span][1]
        expected_sum_left = np.sum(self.left_cum_dist)
        expected_sum_right = np.sum(self.right_cum_dist)
        if n_sum_left != expected_sum_left or n_sum_right != expected_sum_right:
            print(n_sum_left, expected_sum_left, n_sum_right, expected_sum_right)
            print(self.left_cum_dist)
            print(self.right_cum_dist)
            print(self.values)
            print(self.spans)
            print(self.cur_solution)
            assert False

    @staticmethod
    def energy_from_values(values: Sequence[str]) -> float:
        """Compute the energy of a sequence of strings.

        Independent of the current optimized assignment.

        Parameters
        ----------
        values:
            Values to compute the energy of.

        Returns
        -------
        float:
            Computed energy.
        """
        length_array = np.array([len(v) for v in values])
        lengths, counts = np.unique(length_array, return_counts=True)
        cum_len = np.zeros(np.max(lengths))

        # Compute the cumulative lengths.
        for i_len, cur_len in enumerate(lengths):
            cur_counts = counts[i_len]
            cum_len[:cur_len] += cur_counts
        return np.sum(np.log(cum_len + 1))

    def set_span(self, i_val: int, i_span: int) -> None:
        """Change the assignment for a value.

        Parameters
        ----------
        i_val:
            Index of the value to be reassigned.
        i_span:
            Index of the span for that value to be assigned to.
        """
        old_span = self.spans[i_val][self.cur_solution[i_val]]
        new_span = self.spans[i_val][i_span]

        # First change the ones on the left
        if old_span[0] > new_span[0]:
            self.left_cum_dist[new_span[0]:old_span[0]] -= 1
        else:
            self.left_cum_dist[old_span[0]:new_span[0]] += 1

        # Then on the right.
        old_right, new_right = len(self.values[i_val]) - np.array([old_span[1], new_span[1]])
        if old_right > new_right:
            self.right_cum_dist[new_right:old_right] -= 1
        else:
            self.right_cum_dist[old_right:new_right] += 1

        # Update the current solution.
        self.cur_solution[i_val] = i_span

    @property
    def new_values(self) -> Tuple[List[str], List[str]]:
        """Get the values on the left and right with the current assignment."""
        left_values = []
        right_values = []
        for i_val, val in enumerate(self.values):
            if self.cur_solution[i_val] == SIDE_LEFT:
                left_values.append(val)
                right_values.append("")
            elif self.cur_solution[i_val] == SIDE_RIGHT:
                left_values.append("")
                right_values.append(val)
            else:
                span = self.spans[i_val][self.cur_solution[i_val]]
                left_values.append(val[:span[0]])
                right_values.append(val[span[1]:])
        return left_values, right_values

    @property
    def statistics(self) -> Dict[str, Any]:
        """Get the minimum/maximum length of the substituted regex and fraction of assignments."""
        # Compute lengths for which the regex substitutes.
        match_lengths: List[int] = []
        for i_val in range(len(self.values)):
            if len(self.spans[i_val]) > 0:
                span = self.spans[i_val][self.cur_solution[i_val]]
                match_lengths.append(span[1]-span[0])

        if len(match_lengths) == 0:
            return {
                "min_len": 1,
                "max_len": 1,
                "frac_used": 0,
                "digits_used": np.zeros(2, dtype=int),
                "n_values": len(self.values),
            }

        min_len = np.min(match_lengths)
        max_len = np.max(match_lengths)
        frac_used = len(match_lengths)/len(self.values)
        digits_used = np.zeros(max_len+1, dtype=int)
        length, counts = np.unique(match_lengths, return_counts=True)
        for i_len, cur_len in enumerate(length):
            digits_used[cur_len] = counts[i_len]
        digits_used[0] = len(self.values)-len(match_lengths)
        # digits_used = digits_used/np.sum(digits_used)
        return {
            "min_len": min_len,
            "max_len": max_len,
            "frac_used": frac_used,
            "digits_used": digits_used,
            "n_values": len(self.values),
        }



"""Testing utilities for plugins."""


from __future__ import annotations

import jsonschema
import polars as pl
from jsonschema.exceptions import SchemaError

from metasynth.dataset import _jsonify
from metasynth.distribution.base import BaseDistribution
from metasynth.privacy import BasePrivacy
from metasynth.provider import (get_distribution_provider,
                                BaseDistributionProvider,
                                DistributionProviderList)


def check_distribution_provider(provider_name: str):
    """Check internal consistency of a distribution provider.

    Arguments
    ---------
    provider_name:
        Name of the provider to be tested.
    """
    provider = get_distribution_provider(provider_name)
    print(type(provider))
    assert isinstance(provider, BaseDistributionProvider)
    assert len(provider.distributions) > 0
    assert all(issubclass(dist, BaseDistribution) for dist in provider.distributions)
    assert isinstance(provider.name, str)
    assert len(provider.name) > 0
    assert provider.name == provider_name
    assert isinstance(provider.version, str)
    assert len(provider.version) > 0


def check_distribution(distribution: type[BaseDistribution], privacy: BasePrivacy,
                       provenance: str):
    """Check whether the distributions in the package can be validated positively.

    Arguments
    ---------
    distribution:
        Distribution to validate to check whether it behaves as expected.
    privacy:
        Level/type of privacy the distribution adheres to.
    provenance:
        Which provider/plugin/package provides the distribution.
    """
    # Check the schema of the distribution.
    schema = distribution.schema()
    dist_dict = distribution.default_distribution().to_dict()
    try:
        jsonschema.validate(_jsonify(dist_dict), schema)
    except SchemaError as err:
        raise ValueError(f"Failed distribution validation for {distribution.__name__}") from err

    # Check the privacy
    assert privacy.is_compatible(distribution)
    DistributionProviderList(provenance).find_distribution(distribution.implements, privacy)

    assert len(distribution.implements.split(".")) == 2
    assert distribution.provenance == provenance
    assert distribution.var_type != "unknown"
    dist = distribution.default_distribution()
    series = pl.Series([dist.draw() for _ in range(100)])
    new_dist = distribution.fit(series, **privacy.fit_kwargs)
    assert isinstance(new_dist, distribution)
    assert set(list(new_dist.to_dict())) >= set(
        ("implements", "provenance", "class_name", "parameters"))



"""Module for distribution providers.

These are used to find/fit distributions that are available. See pyproject.toml on how the
builtin distribution provider is registered.
"""

from __future__ import annotations

import inspect
import warnings
from abc import ABC
from typing import Any, List, Optional, Type, Union

try:
    from importlib_metadata import entry_points, EntryPoint
except ImportError:
    from importlib.metadata import entry_points, EntryPoint  # type: ignore

import numpy as np
import polars as pl

from metasynth.distribution.na import NADistribution
from metasynth.distribution.base import BaseDistribution
from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.continuous import (ExponentialDistribution,
                                               LogNormalDistribution,
                                               NormalDistribution,
                                               TruncatedNormalDistribution,
                                               UniformDistribution)
from metasynth.distribution.datetime import (UniformDateDistribution,
                                             UniformDateTimeDistribution,
                                             UniformTimeDistribution)
from metasynth.distribution.discrete import (DiscreteUniformDistribution,
                                             PoissonDistribution,
                                             UniqueKeyDistribution)
from metasynth.distribution.faker import (FakerDistribution,
                                          UniqueFakerDistribution,
                                          FreeTextDistribution)
from metasynth.distribution.regex import (RegexDistribution,
                                          UniqueRegexDistribution)
from metasynth.distribution import legacy
from metasynth.privacy import BasePrivacy, BasicPrivacy


class BaseDistributionProvider(ABC):
    """Class that encapsulates a set of distributions.

    It has a property {var_type}_distributions for every var_type.
    """

    name = ""
    version = ""
    distributions: list[type[BaseDistribution]] = []
    legacy_distributions: list[type[BaseDistribution]] = []

    def __init__(self):
        # Perform internal consistency check.
        assert len(self.name) > 0
        assert len(self.version) > 0
        assert len(self.distributions) > 0

    def get_dist_list(self, var_type: str,
                      use_legacy: bool = False) -> List[Type[BaseDistribution]]:
        """Get all distributions for a certain variable type.

        Parameters
        ----------
        var_type:
            Variable type to get the distributions for.
        use_legacy:
            Whether to find the distributions in the legacy distribution list.

        Returns
        -------
        list[Type[BaseDistribution]]:
            List of distributions with that variable type.
        """
        dist_list = []
        if use_legacy:
            distributions = self.legacy_distributions
        else:
            distributions = self.distributions
        for dist_class in distributions:
            str_chk = (isinstance(dist_class.var_type, str) and var_type == dist_class.var_type)
            lst_chk = (not isinstance(dist_class.var_type, str) and var_type in dist_class.var_type)
            if str_chk or lst_chk:
                dist_list.append(dist_class)
        return dist_list

    @property
    def all_var_types(self) -> List[str]:
        """Return list of available variable types."""
        var_type_set = set()
        for dist in self.distributions:
            if isinstance(dist.var_type, str):
                var_type_set.add(dist.var_type)
            else:
                var_type_set.update(dist.var_type)
        return list(var_type_set)


class BuiltinDistributionProvider(BaseDistributionProvider):
    """Distribution tree that includes the builtin distributions."""

    name = "builtin"
    version = "1.1"
    distributions = [
        DiscreteUniformDistribution, PoissonDistribution, UniqueKeyDistribution,
        UniformDistribution, NormalDistribution, LogNormalDistribution,
        TruncatedNormalDistribution, ExponentialDistribution,
        MultinoulliDistribution,
        RegexDistribution, UniqueRegexDistribution, FakerDistribution, UniqueFakerDistribution,
        FreeTextDistribution,
        UniformDateDistribution,
        UniformTimeDistribution,
        UniformDateTimeDistribution,
    ]
    legacy_distributions = [
        legacy.RegexDistribution, legacy.UniqueRegexDistribution
    ]


class DistributionProviderList():
    """List of DistributionProviders with functionality to fit distributions.

    Arguments
    ---------
    dist_providers:
        One or more distribution providers, that are denoted either with a string ("builtin")
        , DistributionProvider (BuiltinDistributionProvider()) or DistributionProvider type
        (BuiltinDistributionProvider).
        The order in which distribution providers are included matters. If a provider implements
        the same distribution at the same privacy level, then only the first will be taken into
        account.
    """

    def __init__(
            self,
            dist_providers: Union[
                None, str, type[BaseDistributionProvider], BaseDistributionProvider,
                list[Union[str, type[BaseDistributionProvider], BaseDistributionProvider]]]):
        if dist_providers is None:
            self.dist_packages = _get_all_provider_list()
            return

        if isinstance(dist_providers, (str, type, BaseDistributionProvider)):
            dist_providers = [dist_providers]
        self.dist_packages = []
        for provider in dist_providers:
            if isinstance(provider, str):
                self.dist_packages.append(get_distribution_provider(provider))
            elif isinstance(provider, type):
                self.dist_packages.append(provider())
            elif isinstance(provider, BaseDistributionProvider):
                self.dist_packages.append(provider)
            else:
                raise ValueError(f"Unknown distribution package type '{type(provider)}'")

    def fit(self, series: pl.Series,
            var_type: str,
            dist: Optional[Union[str, BaseDistribution, type]] = None,
            privacy: BasePrivacy = BasicPrivacy(),
            unique: Optional[bool] = None,
            fit_kwargs: Optional[dict] = None):
        """Fit a distribution to a column/series.

        Parameters
        ----------
        series:
            The data to fit the distributions to.
        var_type:
            The variable type of the data.
        dist:
            Distribution to fit. If not supplied or None, the AIC information
            criterion will be used to determine which distribution is the most
            suitable.
        privacy:
            Level of privacy that will be used in the fit.
        unique:
            Whether the distribution should be unique or not.
        fit_kwargs:
            Extra options for distributions during the fitting stage.
        """
        if fit_kwargs is None:
            fit_kwargs = {}
        if dist is not None:
            return self._fit_distribution(series, dist, privacy, **fit_kwargs)
        if len(fit_kwargs) > 0:
            raise ValueError(f"Got fit arguments for variable '{series.name}', but no "
                             "distribution. Set the distribution manually to fix.")
        return self._find_best_fit(series, var_type, unique, privacy)

    def _find_best_fit(self, series: pl.Series, var_type: str,
                       unique: Optional[bool],
                       privacy: BasePrivacy) -> BaseDistribution:
        """Fit a distribution to a series.

        Search for the distribution within all available distributions in the tree.

        Parameters
        ----------
        series:
            Series to fit a distribution to.
        var_type:
            Variable type of the series.
        unique:
            Whether the variable should be unique or not.
        privacy:
            Privacy level to find the best fit with.

        Returns
        -------
        BaseDistribution:
            Distribution fitted to the series.
        """
        if len(series.drop_nulls()) == 0:
            return NADistribution()
        dist_list = self._get_dist_list(privacy, var_type)
        if len(dist_list) == 0:
            raise ValueError(f"No available distributions with variable type: '{var_type}'")
        dist_instances = [d.fit(series, **privacy.fit_kwargs) for d in dist_list]
        dist_aic = [d.information_criterion(series) for d in dist_instances]
        i_best_dist = np.argmin(dist_aic)
        warnings.simplefilter("always")
        if dist_instances[i_best_dist].is_unique and unique is None:
            warnings.warn(f"\nVariable {series.name} seems unique, but not set to be unique.\n"
                          "Set the variable to be either unique or not unique to remove this "
                          "warning.\n")
        if unique is None:
            unique = False

        dist_aic = [dist_aic[i] for i in range(len(dist_aic))
                    if dist_instances[i].is_unique == unique]
        dist_instances = [d for d in dist_instances if d.is_unique == unique]
        if len(dist_instances) == 0:
            raise ValueError(f"No available distributions for variable '{series.name}'"
                             f" with variable type '{var_type}' "
                             f"that have unique == {unique}.")
        return dist_instances[np.argmin(dist_aic)]

    def find_distribution(self,  # pylint: disable=too-many-branches
                          dist_name: str,
                          privacy: BasePrivacy = BasicPrivacy(),
                          var_type: Optional[str] = None,
                          version: Optional[str] = None) -> type[BaseDistribution]:
        """Find a distribution and fit keyword arguments from a name.

        Parameters
        ----------
        dist_name:
            Name of the distribution, e.g., for the built-in
            uniform distribution: "uniform", "core.uniform", "UniformDistribution".
        privacy:
            Type of privacy to be applied.
        var_type:
            Type of the variable to find. If var_type is None, then do not check the
            variable type.
        version:
            Version of the distribution to get. If necessary get them from legacy.

        Returns
        -------
        tuple[Type[BaseDistribution], dict[str, Any]]:
            A distribution and the arguments to create an instance.
        """
        if NADistribution.matches_name(dist_name):
            return NADistribution

        versions_found = []
        for dist_class in self._get_dist_list(privacy, var_type=var_type) + [NADistribution]:
            if dist_class.matches_name(dist_name):
                if version is None or version == dist_class.version:
                    return dist_class
                versions_found.append(dist_class)

        # Look for distribution in legacy
        warnings.simplefilter("always")
        legacy_versions: list[Type[BaseDistribution]] = []
        for dist_class in self._get_dist_list(privacy, use_legacy=True, var_type=var_type):
            if dist_class.matches_name(dist_name):
                if version is None or version == dist_class.version:
                    if len(versions_found) == 0:
                        warnings.warn("Distribution with name '{dist_name}' is deprecated and "
                                      "will be removed in the future.")
                        return dist_class
                    warnings.warn("Version ({version}) of distribution with name '{dist_name}'"
                                  " is deprecated and will be removed in the future.")
        if version is not None:
            major_version = version.split(".")[0]
            for dist_class in versions_found:
                if dist_class.version.split(".")[0] == major_version:
                    warnings.warn("Version mismatch ({version}) versus ({dist_class.version})")
                    return dist_class
            for dist_class in legacy_versions:
                if dist_class.version.split(".")[0] == major_version:
                    warnings.warn("Version mismatch ({version}) versus ({dist_class.version})"
                                  " for '{dist_name}")
                    warnings.warn("(Major) Version ({dist_classversion}) of distribution is"
                                  " deprecated")
                    return dist_class
        if len(legacy_versions+versions_found) == 0:
            raise ValueError(f"Cannot find distribution with name '{dist_name}'.")
        raise ValueError(
            f"Cannot find compatible version for distribution '{dist_name}', available: "
            f"{legacy_versions+versions_found}")

    def _fit_distribution(self, series: pl.Series,
                          dist: Union[str, Type[BaseDistribution], BaseDistribution],
                          privacy: BasePrivacy,
                          **fit_kwargs) -> BaseDistribution:
        """Fit a specific distribution to a series.

        In contrast the fit method, this needs a supplied distribution(type).

        Parameters
        ----------
        dist:
            Distribution to fit (if it is not already fitted).
        series:
            Series to fit the distribution to.
        privacy:
            Privacy level to fit the distribution with.
        fit_kwargs:
            Extra keyword arguments to modify the way the distribution is fit.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        dist_instance = None
        if isinstance(dist, BaseDistribution):
            return dist

        if isinstance(dist, str):
            dist_class = self.find_distribution(dist, privacy=privacy)
        elif inspect.isclass(dist) and issubclass(dist, BaseDistribution):
            dist_class = dist
        else:
            raise TypeError(
                f"Distribution {dist} with type {type(dist)} is not a BaseDistribution")
        if issubclass(dist_class, NADistribution):
            dist_instance = dist_class.default_distribution()
        else:
            dist_instance = dist_class.fit(series, **privacy.fit_kwargs, **fit_kwargs)
        return dist_instance

    def _get_dist_list(self, privacy: Optional[BasePrivacy] = None,
                       var_type: Optional[str] = None,
                       use_legacy: bool = False) -> list[type[BaseDistribution]]:
        dist_list = []
        for dist_provider in self.dist_packages:
            if var_type is None:
                dist_list.extend(dist_provider.distributions)
            else:
                dist_list.extend(dist_provider.get_dist_list(var_type,
                                                             use_legacy=use_legacy))

        if privacy is None:
            return dist_list
        dist_list = [dist for dist in dist_list if dist.privacy == privacy.name]
        return dist_list

    def from_dict(self, var_dict: dict[str, Any]) -> BaseDistribution:
        """Create a distribution from a dictionary.

        Parameters
        ----------
        var_dict:
            Variable dictionary that includes the distribution properties.

        Returns
        -------
        BaseDistribution:
            Distribution representing the dictionary.
        """
        dist_name = var_dict["distribution"]["implements"]
        version = var_dict["distribution"].get("version", "1.0")
        var_type = var_dict["type"]
        dist_class = self.find_distribution(dist_name, version=version, var_type=var_type)
        return dist_class.from_dict(var_dict["distribution"])


def _get_all_providers() -> dict[str, EntryPoint]:
    """Get all available providers."""
    return {
        entry.name: entry
        for entry in entry_points(group="metasynth.distribution_provider")
    }


def _get_all_provider_list() -> list[BaseDistributionProvider]:
    return [p.load()() for p in _get_all_providers().values()]


def get_distribution_provider(
        provider: Union[str, type[BaseDistributionProvider],
                        BaseDistributionProvider] = "builtin"
        ) -> BaseDistributionProvider:
    """Get a distribution tree.

    Parameters
    ----------
    provider:
        Name, class or class type of the provider to be used.
    kwargs:
        Extra keyword arguments for initialization of the distribution provider.

    Returns
    -------
    BaseDistributionProvider:
        The distribution provider that was found.
    """
    if isinstance(provider, BaseDistributionProvider):
        return provider
    if isinstance(provider, type):
        return provider()

    all_providers = _get_all_providers()
    try:
        return all_providers[provider].load()()
    except KeyError as exc:
        raise ValueError(f"Cannot find distribution provider with name '{provider}'.") from exc



import numpy as np
from scipy import stats
from metasynth.distribution.continuous import UniformDistribution,\
    NormalDistribution, LogNormalDistribution, TruncatedNormalDistribution,\
    ExponentialDistribution
from pytest import mark


@mark.parametrize(
    "lower_bound,upper_bound",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_uniform(lower_bound, upper_bound):
    scale = upper_bound-lower_bound
    values = stats.uniform(loc=lower_bound, scale=scale).rvs(100)
    dist = UniformDistribution.fit(values)
    assert dist.min_val <= values.min()
    assert dist.max_val >= values.max()
    assert dist.information_criterion(values) < 4 - 200*np.log((upper_bound-lower_bound)**-1)
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "mean,std_dev",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_normal(mean, std_dev):
    values = stats.norm(loc=mean, scale=std_dev).rvs(1000)
    dist = NormalDistribution.fit(values)
    dist_uniform = UniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mean - mean)/std_dev < 0.5
    assert (dist.std_dev - std_dev)/std_dev < 0.5
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "mu,sigma",
    [
        (0, 1),
        (-10, 2),
        (0.01, 0.02),
    ]
)
def test_log_normal(mu, sigma):
    values = stats.lognorm(s=sigma, loc=0, scale=np.exp(mu)).rvs(1000)
    dist = LogNormalDistribution.fit(values)
    dist_uniform = UniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert (dist.mu-mu)/sigma < 1
    assert (dist.sigma-sigma)/sigma < 1
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "lower_bound,upper_bound,mu,sigma",
    [
        (-0.5, 0.5, 0, 0.5),
        (-10, -8, 0, 5),
        (-5, 3, 2, 2),
        (-0.01, 0.01, 0.01, 0.01),
    ]
)
def test_trunc_normal(lower_bound, upper_bound, mu, sigma):
    a, b = (lower_bound-mu)/sigma, (upper_bound-mu)/sigma
    values = stats.truncnorm(a=a, b=b, loc=mu, scale=sigma).rvs(5000)
    dist = TruncatedNormalDistribution.fit(values)
    dist_uniform = UniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), float)


@mark.parametrize(
    "rate",
    [0.1, 10, 100, 234.1234]
)
def test_exponential(rate):
    values = stats.expon(loc=0, scale=1/rate).rvs(5000)
    dist = ExponentialDistribution.fit(values)
    dist_uniform = UniformDistribution.fit(values)
    assert dist.information_criterion(values) < dist_uniform.information_criterion(values)
    assert isinstance(dist.draw(), float)
    assert (dist.rate - rate)/rate < 0.1



"""Package for fitting regexes to strings."""

from metasynth.distribution.legacy.regex.base import (RegexDistribution,
                                                      UniqueRegexDistribution)

__all__ = ["RegexDistribution", "UniqueRegexDistribution"]



"""Module containing an interface to the faker package."""
from typing import Iterable, Optional

from faker import Faker
from lingua import LanguageDetectorBuilder
from lingua._constant import LETTERS, PUNCTUATION
# LETTERS: Pattern = regex.compile(r"\p{Han}|\p{Hangul}|\p{Hiragana}|\p{Katakana}|\p{L}+")
# PUNCTUATION: Pattern = regex.compile(r"\p{P}")
from scipy.stats import poisson

from metasynth.distribution.base import metadist, BaseDistribution, UniqueDistributionMixin


@metadist(implements="core.faker", var_type="string")
class FakerDistribution(BaseDistribution):
    """Distribution for the faker package.

    This is mainly an interface for the faker package, so that it
    can be used within the MetaSynth package. It doesn't have any
    true fitting/statistical inference method, so it has to be manually
    selected.

    Parameters
    ----------
    faker_type: str
        The provider function in the faker package, e.g. 'city' or 'ipv4', etc.
    locale: str
        Locale used for the faker package.
    """

    def __init__(self, faker_type: str, locale: str = "en_US"):
        self.faker_type: str = faker_type
        self.locale: str = locale
        self.fake: Faker = Faker(locale=locale)

    @classmethod
    def _fit(cls, values, faker_type: str = "city", locale: str = "en_US"):  \
            # pylint: disable=arguments-differ
        """Select the appropriate faker function and locale."""
        return cls(faker_type, locale)

    def draw(self):
        return getattr(self.fake, self.faker_type)()

    def information_criterion(self, values: Iterable) -> float:
        return 99999

    @classmethod
    def default_distribution(cls):
        return cls("city")

    def _param_dict(self):
        return {
            "faker_type": self.faker_type,
            "locale": self.locale
        }

    @classmethod
    def _param_schema(cls):
        return {
            "faker_type": {"type": "string"},
            "locale": {"type": "string"},
        }


@metadist(implements="core.unique_faker", var_type="string")
class UniqueFakerDistribution(UniqueDistributionMixin, FakerDistribution):
    """Faker distribution that returns unique values."""


@metadist(implements="core.freetext", var_type="string")
class FreeTextDistribution(BaseDistribution):
    """Distribution for free text.

    This distribution detects the language and generates sentences using
    the Faker package. The average number of sentences and words per item
    are detected using regexes (with the lingua package).

    Parameters
    ----------
    locale: str
        Locale used for the faker package.
    avg_sentences:
        Average number of sentences (punctuation marks) per (non-NA) row,
        if None do not make sentences.
    avg_words:
        Average number of words per (non-NA) row.
    """

    def __init__(self, locale: str, avg_sentences: Optional[float], avg_words: int):
        self.locale: str = locale
        self.avg_sentences = avg_sentences
        self.avg_words = avg_words
        self.fake = Faker(locale=self.locale)

    @classmethod
    def _fit(cls, values):
        """Select the appropriate faker function and locale."""
        lang_str = cls.detect_language(values)
        if lang_str is None:
            return cls.default_distribution()

        try:
            Faker(lang_str)
        except AttributeError:
            lang_str = "EN"

        all_text = "\n".join(values)
        n_non_empty = (values != "").sum()
        n_punctuation = len(list(PUNCTUATION.finditer(all_text)))
        n_words = len(list(LETTERS.finditer(all_text)))
        if n_punctuation < n_non_empty//3:
            avg_sentence = None
        else:
            avg_sentence = n_punctuation/len(values)
        avg_words = n_words/len(values)
        return cls(lang_str, avg_sentence, avg_words)

    @classmethod
    def detect_language(cls, values: Iterable) -> Optional[str]:
        """Detect the language of some text.

        Parameters
        ----------
        values:
            Values to detect the language of (usually polars dataframe).

        Returns
        -------
        language:
            Two letter ISO code to represent the language, or None if it could not be determined.
        """
        detector = LanguageDetectorBuilder.from_all_languages().with_low_accuracy_mode().build()
        lang = detector.detect_language_of("\n".join(values))
        if lang is None:
            return None
        return str(lang.iso_code_639_1).rsplit(".", maxsplit=1)[-1]

    def draw(self):
        if self.avg_sentences is None:
            n_words = max(1, poisson(self.avg_words).rvs())
            sentence = self.fake.sentence(n_words)
            return sentence[:-1]

        n_sentences = max(1, poisson(self.avg_sentences).rvs())
        avg_words_per_sent = max(1, self.avg_words/max(1, self.avg_sentences))
        n_words = max(1, poisson(avg_words_per_sent).rvs())
        return " ".join(self.fake.sentence(nb_words=n_words) for _ in range(n_sentences))

    def information_criterion(self, values) -> float:
        # series = self._to_series(values)
        # lang = self.detect_language(series)
        # if lang is None:
        # Don't use this distribution by default (for now).
        return 99999999

    @classmethod
    def default_distribution(cls):
        return cls("en_US", 3, 6)

    def _param_dict(self):
        return {
            "locale": self.locale,
            "avg_sentences": self.avg_sentences,
            "avg_words": self.avg_words,
        }

    @classmethod
    def _param_schema(cls):
        return {
            "locale": {"type": "string"},
            "avg_sentences": {"type": ["number", "null"]},
            "avg_words": {"type": "number"},
        }



from pytest import mark

from metasynth.provider import get_distribution_provider
from metasynth.testutils import check_distribution, check_distribution_provider
from metasynth.privacy import BasicPrivacy


def test_builtin_provider():
    check_distribution_provider("builtin")


@mark.parametrize(
    "distribution", get_distribution_provider("builtin").distributions
)
def test_dist_validation(distribution):
    check_distribution(distribution, privacy=BasicPrivacy(),
                       provenance="builtin")



"""Module containing string distributions."""

from __future__ import annotations

from typing import List, Sequence, Set, Tuple, Type, Union

import numpy as np

from metasynth.distribution.base import metadist, BaseDistribution
from metasynth.distribution.legacy.regex.element import (AlphaNumericRegex, AnyRegex,
                                                         BaseRegexElement, DigitRegex,
                                                         LettersRegex, LowercaseRegex,
                                                         SingleRegex, UppercaseRegex)
from metasynth.distribution.legacy.regex.optimizer import RegexOptimizer


def _get_gradient_start(values: Sequence[str], new_values: Sequence[str],
                        regex_elem: BaseRegexElement) -> float:
    """Get the proportion of the characters that were resolved with the regex."""
    digits_used: List[int] = []
    for i_val, old_val in enumerate(values):
        new_val = new_values[i_val]
        d_len = len(old_val) - len(new_val)
        if len(digits_used) <= d_len:
            digits_used += (d_len - len(digits_used) + 1)*[0]
        digits_used[d_len] += 1
    regex_stat = {
        "n_values": len(values),
        "frac_used": regex_elem.frac_used,
        "digits_used": digits_used,
    }
    old_energy = RegexOptimizer.energy_from_values(values)
    new_energy = RegexOptimizer.energy_from_values(new_values)
    energy_budget = regex_elem.information_budget(regex_stat)

    delta_energy = old_energy - new_energy
    return delta_energy/energy_budget


@metadist(implements="core.regex", var_type="string")
class RegexDistribution(BaseDistribution):
    """Distribution that uses a strategy similar to regex.

    The idea behind this method is that for structured strings
    regexes works very well. It uses a greedy algorithm to build the
    regex. It does not implement all regexes, nor does it claim to be
    compliant with standard ones.

    Parameters
    ----------
    re_list: list of BaseRegexElement
        List of basic regex elements in the order that they occur.
    """

    implements = "core.regex"

    def __init__(self, re_list: Union[Sequence[Union[BaseRegexElement, Tuple[str, float]]], str]):
        self.re_list = []
        if isinstance(re_list, str):
            self.re_list = self._unpack_regex(re_list)
            return

        for re_elem in re_list:
            if isinstance(re_elem, BaseRegexElement):
                self.re_list.append(re_elem)
                continue
            regex_str, frac_used = re_elem
            regex_return = None
            for regex_class in self.all_regex_classes():
                regex_return = regex_class.from_string(regex_str, frac_used)
                if regex_return is not None and regex_return[1] == "":
                    break
            if regex_return is None:
                raise ValueError(f"Unrecognized regex element '{regex_str}'")
            regex_dist_elem = regex_return[0]
            self.re_list.append(regex_dist_elem)

    @classmethod
    def all_regex_classes(cls) -> List[Type[BaseRegexElement]]:
        """Return all regex element classes."""
        return [
            DigitRegex, AlphaNumericRegex, LowercaseRegex, UppercaseRegex,
            LettersRegex, SingleRegex, AnyRegex,
        ]

    def _unpack_regex(self, regex_str: str):
        if len(regex_str) == 0:
            return []
        for re_elem_class in self.all_regex_classes():
            ret = re_elem_class.from_string(regex_str)
            if ret is not None:
                return [ret[0]] + self._unpack_regex(ret[1])
        raise ValueError("Failed to determine regex from '" + regex_str + "'")

    @classmethod
    def _fit(cls, values, mode: str = "fast"):
        if mode == "fast":
            return cls._fit_fast(values)
        return cls._fit_slow(values)

    @classmethod
    def _fit_slow(cls, values):
        if np.sum([len(val) for val in values]) == 0:
            return cls([])

        best_solution = None
        regex_classes = cls.all_regex_classes()

        # Iterate over all RegexElements and find the best one.
        for re_class in regex_classes:
            new_values, gradient, regexer = re_class.fit(list(values))
            if best_solution is None or gradient > best_solution["gradient"]:
                best_solution = {
                    "re": regexer,
                    "gradient": gradient,
                    "values": new_values,
                }
        if best_solution is None:
            return cls([])
        left_fit = cls._fit_slow(best_solution["values"][0])
        middle_fit = cls([best_solution["re"]])
        right_fit = cls._fit_slow(best_solution["values"][1])
        return left_fit + middle_fit + right_fit

    @classmethod
    def _fit_fast(cls, values):
        n_char = np.sum([len(val) for val in values])
        if n_char == 0:
            return cls([])
        best_solution = None
        regex_classes = cls.all_regex_classes()

        # Iterate over all RegexElements and find the best one.
        for re_class in regex_classes:
            new_values, regexer = re_class.fit_start(list(values))
            # n_char_removed = _get_n_char_removed(values, new_values)
            # cur_ic = regexer.information_criterion(n_char_removed)
            # if cur_ic > 1.5:
            # continue
            gradient = _get_gradient_start(values, new_values, regexer)
            if best_solution is None or gradient > best_solution["gradient"]:
                best_solution = {
                    "gradient": gradient,
                    "re": regexer,
                    "values": new_values,
                }
        if best_solution is None:
            return cls([])
        return cls([best_solution["re"]]) + cls._fit_fast(best_solution["values"])

    def __add__(self, other: RegexDistribution) -> RegexDistribution:
        return self.__class__(self.re_list + other.re_list)

    def draw(self):
        cur_str = ""
        for rex in self.re_list:
            cur_str += rex.draw()
        return cur_str

    @property
    def regex_string(self) -> str:
        """String representation of the string."""
        return "".join([str(x) for x in self.re_list])

    def _param_dict(self):
        return {
            "re_list": [(str(x), x.frac_used) for x in self.re_list],
        }

    @classmethod
    def _param_schema(cls):
        return {
            "re_list": {
                "type": "array",
                "items": {
                    "type": "array",
                    "prefixItems": [{"type": "string"}, {"type": "number"}],
                    "minItems": 2,
                    "additionalItems": False
                }
            }
        }

    @classmethod
    def default_distribution(cls):
        return cls([(r"\d{3,4}", 0.67)])


@metadist(implements="core.unique_regex", var_type="string", is_unique=True)
class UniqueRegexDistribution(RegexDistribution):
    """Unique variant of the regex distribution.

    Same as the normal regex distribution, but checks whether a key
    has already been used.

    Parameters
    ----------
    re_list: list of BaseRegexElement
        List of basic regex elements in the order that they occur.
    """

    def __init__(self, re_list: Sequence[Union[BaseRegexElement, Tuple[str, float]]]):
        super().__init__(re_list)
        self.key_set: Set[str] = set()

    def draw_reset(self):
        self.key_set = set()

    def draw(self) -> str:
        n_try = 0
        while n_try < 1e5:
            new_val = super().draw()
            if new_val not in self.key_set:
                self.key_set.add(new_val)
                return new_val
            n_try += 1
        raise ValueError("Failed to draw unique string after 100.000 tries.")

    def information_criterion(self, values):
        return 99999



"""Module containing categorical distributions."""

from __future__ import annotations

from typing import Union
import warnings

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl

from metasynth.distribution.base import metadist, BaseDistribution


@metadist(implements="core.multinoulli", var_type=["categorical", "discrete", "string"])
class MultinoulliDistribution(BaseDistribution):
    """Categorical distribution that stores category labels and probabilities.

    Parameters
    ----------
    labels: list of str
        List containing the label belonging to each category.
    probs: list of int
        List containing the probability of each category.
        Probabilities will be normalized, so frequencies are valid too.
    """

    def __init__(self, labels: Union[npt.NDArray[np.str_], list[str]],
                 probs: Union[npt.NDArray[np.float_], list[float]]):
        self.labels = np.array(labels)
        self.probs = np.array(probs)
        if not np.isclose(np.sum(self.probs), 1):
            if np.any(self.probs < 0):
                raise ValueError("Cannot create multinoulli distribution with probabilities < 0.")
            warnings.simplefilter("always")
            warnings.warn("Multinoulli probabilities do not add up to 1 "
                          f" ({np.sum(self.probs)}); they will be rescaled.")
        self.probs = self.probs/np.sum(self.probs)

    @classmethod
    def _fit(cls, values: pl.Series):
        labels, counts = np.unique(values, return_counts=True)
        probs = counts/np.sum(counts)
        return cls(labels, probs)

    def _param_dict(self):
        return {"labels": self.labels, "probs": self.probs}

    @classmethod
    def _param_schema(cls):
        return {
            "labels": {"type": "array", "uniqueItems": True},
            "probs": {"type": "array", "items": {"type": "number"}},
        }

    def draw(self):
        return np.random.choice(self.labels, p=self.probs)

    def information_criterion(self,
                              values: Union[pd.Series, pl.Series, npt.NDArray[np.str_]]
                              ) -> float:
        series = self._to_series(values)
        labels, counts = np.unique(series, return_counts=True)
        log_lik = 0.0
        pdict = dict(zip(self.labels, self.probs))
        # Check type of variable and act accordingly.
        if len(self.labels) > 1 and isinstance(self.labels[0], (int, np.int_)):
            log_lik = self._log_like_int(series, labels, counts)
            # See docstring of log_like_int method.
            n_parameters = len(self.probs)-1 + len(self.probs)
        else:
            for lab, count in zip(labels, counts):
                # account for missing values / missing categories
                # by setting default of .get to 1 (add log(1)=0 to log_lik)
                log_lik += count * np.log(pdict.get(lab, 1))
            n_parameters = len(self.probs)-1

        return 2*n_parameters - 2*log_lik

    def _log_like_int(
            self,
            series: pl.Series,
            labels: npt.NDArray[np.int_],
            counts: npt.NDArray[np.int_],
            alpha: float = 1.0,
            ) -> float:
        """Information criterion for the integer variant of the multinoulli distribution.

        This information criterion is a little bit different from the AIC, which is why
        the method is split into its own method. The reason is that with the default IC,
        the multinomial distribution is unfairly favored compared to the uniform distribution
        in particular (but really any discrete distribution). Assume for instance that we have a
        uniform distribution [0, 1000], with 10 values, then the log likelihood of the uniform
        distribution would be 10*log(1/1000), while the multinoulli has 10*log(1/10), which is
        much higher. To solve this problem we use additive smoothing:

        https://en.wikipedia.org/wiki/Additive_smoothing

        with parameter alpha = 1. This effectively says that all values in between the observed
        values have a probability of 1/N, where N is the number of total values. Hoewever, to
        prevent confusion (and potentially improve the synthesis) is draw from these unobserved
        values. That is why the information criterion for integers is not a true AIC, but a pseudo
        information criterion.

        The number of parameters is also in discussion, since we could imagine another multinoulli
        distribution with equal probs, which would have 'no' parameters. We could also see the
        multinoulli distribution as a dirac delta distribution with probability prefactors. Then it
        is easy to argue that the position of the dirac delta in effect is also a parameter. Thus,
        we choose to add an extra len(self.labels) term to keep the choice of distribution
        consistent.
        """
        # series = self._to_series(values)
        # labels, counts = np.unique(series, return_counts=True)
        log_lik = 0.0
        if len(self.labels) <= 1:
            return 0
        # Do additive smoothing, assume count of values is equal to that value in initial fit.
        num_event = len(series)
        num_tot_cat = np.max(self.labels)-np.min(self.labels)+1
        num_zero_cat = num_tot_cat - len(self.labels)
        theta = (self.probs*num_event+alpha)/(num_event+num_tot_cat*alpha)
        theta_zero = alpha/(num_event+num_tot_cat*alpha)
        theta_sum = np.sum(theta) + num_zero_cat*theta_zero
        theta_dict = dict(zip(self.labels, theta))
        for lab, count in zip(labels, counts):
            cur_theta = theta_dict.get(lab, theta_zero)
            cur_prob = cur_theta/theta_sum
            log_lik += count * np.log(cur_prob)
        return log_lik

    @classmethod
    def default_distribution(cls):
        return cls(["a", "b", "c"], [0.1, 0.3, 0.6])



"""MetaSynth: a package for creating synthetic datasets.

One part concerns the creation of the statistical metadata from the
original data, while the other part creates a synthetic dataset from the
metadata.
"""

from importlib.metadata import version

from metasynth.dataset import MetaFrame
from metasynth.var import MetaVar
from metasynth.demo.dataset import demo_file
from metasynth.distribution.base import metadist

__all__ = ["MetaVar", "MetaFrame", "demo_file", "metadist"]
__version__ = version("metasynth")



import json

import pandas as pd
import polars as pl
import numpy as np
from metasynth.var import MetaVar
from metasynth.distribution import NormalDistribution, RegexDistribution, UniqueRegexDistribution
from metasynth.distribution import DiscreteUniformDistribution
from metasynth.distribution import UniformDistribution
from pytest import mark, raises
from metasynth.dataset import _jsonify
from metasynth.distribution.discrete import UniqueKeyDistribution
from metasynth.distribution.continuous import TruncatedNormalDistribution
from metasynth.distribution.categorical import MultinoulliDistribution


def _series_drop_nans(series):
    if isinstance(series, pl.Series):
        return series.drop_nulls()
    return series.dropna()


def _series_element_classname(series):
    series = _series_drop_nans(series)
    if isinstance(series, pl.Series):
        return series[0].__class__.__name__
    return series.iloc[0].__class__.__name__


def check_var(series, var_type, temp_path):
    def check_similar(series_a, series_b):
        assert isinstance(series_a, (pd.Series, pl.Series))
        assert isinstance(series_b, (pd.Series, pl.Series))
        assert len(series_a) == len(series_b)
        base_type_a = _series_element_classname(series_a)
        base_type_b = _series_element_classname(series_b)
        if type(series_a) == type(series_b):
            assert base_type_a == base_type_b
        assert (len(series_a)-len(_series_drop_nans(series_a)) > 0) == (len(series_b) - len(_series_drop_nans(series_b)) > 0)

    assert isinstance(series, (pd.Series, pl.Series))
    var = MetaVar.detect(series)
    assert isinstance(str(var), str)
    assert "Proportion of Missing Values" in str(var)

    with raises(ValueError):
        var.draw_series(100)
    var.fit()
    new_series = var.draw_series(len(series))
    check_similar(series, new_series)
    assert var.var_type == var_type
    assert var_type in var.distribution.var_type

    new_var = MetaVar.from_dict(var.to_dict())
    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict.update({"type": "unknown"})
        MetaVar.from_dict(var_dict)

    with raises(ValueError):
        var_dict = var.to_dict()
        var_dict["distribution"].update({"implements": "unknown"})
        MetaVar.from_dict(var_dict)

    newer_series = new_var.draw_series(len(series))
    check_similar(newer_series, series)
    with raises(ValueError):
        new_var.fit()

    assert type(new_var) == type(var)
    assert new_var.dtype == var.dtype
    assert var_type == new_var.var_type

    # Write to JSON file and read it back
    tmp_fp = temp_path / "tmp.json"
    with open(tmp_fp, "w") as f:
        json.dump(_jsonify(var.to_dict()), f)

    with open(tmp_fp, "r") as f:
        new_var = MetaVar.from_dict(json.load(f))
    check_similar(series, new_var.draw_series(len(series)))

    assert type(new_var) == type(var)
    assert new_var.dtype == var.dtype
    assert new_var.var_type == var_type

    return new_series


@mark.parametrize(
    "series",
    [
        pd.Series(np.random.choice(["a", "b", "c", None], size=100), dtype="category"),
        pl.Series(np.random.choice(["a", "b", "c", None], size=100).tolist(), dtype=pl.Categorical)
    ]
)
def test_categorical(tmp_path, series):
    # series = pd.Series(np.random.choice(["a", "b", "c", None], size=100), dtype="category")
    new_series = check_var(series, "categorical", tmp_path)
    assert set(_series_drop_nans(series)) == set(np.unique(_series_drop_nans(new_series)))


@mark.parametrize("dtype", ["int8", "int16", "int32", "int64", "int"])
def test_integer(dtype, tmp_path):
    series = pd.Series([np.random.randint(0, 10) for _ in range(100)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10


@mark.parametrize("dtype", ["Int8", "Int16", "Int32", "Int64"])
def test_nullable_integer(dtype, tmp_path):
    series = pd.Series([np.random.randint(0, 10) if np.random.rand() > 0.5 else None
                        for _ in range(100)], dtype=dtype)
    new_series = check_var(series, "discrete", tmp_path)
    assert new_series.min() >= 0
    assert new_series.max() < 10


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_float(tmp_path, series_type):
    np.random.seed(3727442)
    series = series_type([np.random.rand() for _ in range(10000)])
    new_series = check_var(series, "continuous", tmp_path)
    assert new_series.min() > 0
    assert new_series.max() < 1

    series = pd.Series(np.random.randn(1000))
    check_var(series, "continuous", tmp_path)


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series],
)
def test_string(tmp_path, series_type):
    series = series_type(np.random.choice(["a", "b", "c", None], size=100).tolist())
    new_series = check_var(series, "string", tmp_path)
    assert set(np.unique(_series_drop_nans(series))) == set(np.unique(_series_drop_nans(new_series)))


@mark.parametrize(
    "series_type",
    [pl.Series, pd.Series]
)
def test_bool(tmp_path, series_type):
    series = series_type(np.random.choice([True, False], size=100))
    with raises(ValueError):
        check_var(series, "categorical", tmp_path)


@mark.parametrize(
    "prop_missing",
    [-1, -0.1, 1.2],
)
def test_invalid_prop(prop_missing):
    with raises(ValueError):
        MetaVar("continuous")
    with raises(ValueError):
        MetaVar("continuous", prop_missing=prop_missing)


@mark.parametrize(
    "dataframe",
    [
        pd.DataFrame({
            "int": pd.Series([np.random.randint(0, 10) for _ in range(100)]),
            "float": pd.Series([np.random.rand() for _ in range(100)])
        }),
        pl.DataFrame({
            "int": [np.random.randint(0, 10) for _ in range(100)],
            "float": [np.random.rand() for _ in range(100)]
        })
    ]
)
def test_dataframe(dataframe):
    variables = MetaVar.detect(dataframe)
    assert len(variables) == 2
    assert isinstance(variables, list)
    assert variables[0].var_type == "discrete"
    assert variables[1].var_type == "continuous"


@mark.parametrize(
    "series",
    [pd.Series([np.random.rand() for _ in range(5000)]),
     pl.Series([np.random.rand() for _ in range(5000)])]
)
def test_manual_fit(series):
    var = MetaVar.detect(series)
    var.fit()
    assert isinstance(var.distribution, (UniformDistribution, TruncatedNormalDistribution))
    var.fit("normal")
    assert isinstance(var.distribution, NormalDistribution)
    var.fit(UniformDistribution)
    assert isinstance(var.distribution, UniformDistribution)
    var.fit(NormalDistribution(0, 1))
    assert isinstance(var.distribution, NormalDistribution)
    with raises(TypeError):
        var.fit(10)


@mark.parametrize(
    "series",
    [pd.Series([pd.NA for _ in range(10)]),
     pl.Series([None for _ in range(10)])]
)
def test_na_zero(series):
    var = MetaVar.detect(series)
    var.fit()
    assert var.var_type == "continuous"
    assert var.prop_missing == 1.0


@mark.parametrize(
    "series",
    [pd.Series(np.array([pd.NA if i != 0 else 1.0 for i in range(10)])),
     pl.Series([None if i != 0 else 1.0 for i in range(10)])]
)
def test_na_one(series):
    var = MetaVar.detect(series)
    var.fit()
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.9) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.array([np.nan if i < 2 else 0.123*i for i in range(10)])),
     pl.Series([None if i < 2 else 0.123*i for i in range(10)])]
)
def test_na_two(series):
    var = MetaVar.detect(series)
    var.fit()
    assert var.var_type == "continuous"
    assert abs(var.prop_missing-0.8) < 1e7


@mark.parametrize(
    "series",
    [pd.Series(np.random.randint(0, 100000, size=1000)),
     pl.Series(np.random.randint(0, 100000, size=1000))]
)
def test_manual_unique_integer(series):
    var = MetaVar.detect(series)
    var.fit()
    assert isinstance(var.distribution, DiscreteUniformDistribution)
    var.fit(unique=True)
    assert isinstance(var.distribution, UniqueKeyDistribution)


@mark.parametrize(
    "series",
    [
        pd.Series(["x213", "2dh2", "4k2kk"]),
        pl.Series(["x213", "2dh2", "4k2kk"]),
    ]
)
def test_manual_unique_string(series):
    # series = pd.Series(["x213", "2dh2", "4k2kk"])
    var = MetaVar.detect(series)
    var.fit()
    assert isinstance(var.distribution, RegexDistribution)
    var.fit(unique=True)
    assert isinstance(var.distribution, UniqueRegexDistribution)


@mark.parametrize(
    "series",
    [
        pl.Series([2]*100 + [4]*100 + [12394]*100),
        pl.Series([-123]*12 + [123]*12),
        # pl.Series(["2"]*100 + ["5"]*100 + ["123"]*100),
        pl.Series(["2"]*100 + ["5"]*100 + ["123"]*100, dtype=pl.Categorical)
    ]
)
def test_int_multinoulli(series):
    var = MetaVar.detect(series)
    var.fit()
    assert isinstance(var.distribution, MultinoulliDistribution)



import polars as pl
from metasynth import MetaFrame

# example dataframe from polars website
df = pl.DataFrame(
    {
        "ID": [1, 2, 3, 4, 5],
        "fruits": ["banana", "banana", "apple", "apple", "banana"],
        "B": [5, 4, 3, 2, 1],
        "cars": ["beetle", "audi", "beetle", "beetle", "beetle"],
        "optional": [28, 300, None, 2, -30],
    }
)

# convert appropriate columns to categorical
df = df.with_columns([
    pl.col("fruits").cast(pl.Categorical),
    pl.col("cars").cast(pl.Categorical),
])

# set A to unique and B to not unique
spec_dict = {
    "ID": {"unique": True},
    "B": {"unique": False}
}

# create MetaFrame
mf = MetaFrame.fit_dataframe(df, spec=spec_dict)

# write to json
mf.to_json("examples/basic_example.json")

# then, export json from secure environment

# outside secure environment, load json
mf_out = MetaFrame.from_json("examples/basic_example.json")

# create a fake dataset
mf_out.synthesize(10)


import sys
import subprocess
from pathlib import Path
from pytest import mark, fixture
import polars as pl
from metasynth import MetaFrame

TMP_DIR_PATH = None

@fixture(scope="module")
def tmp_dir(tmp_path_factory) -> Path:
    global TMP_DIR_PATH
    if TMP_DIR_PATH is None:
        # Create a temporary input file
        TMP_DIR_PATH = tmp_path_factory.mktemp("data")
        json_path = TMP_DIR_PATH / "titanic.json"
        csv_fp = Path("tests", "data", "titanic.csv")
        csv_dt = {
            "PassengerId": pl.Int64,
            "Survived": pl.Categorical,
            "Pclass": pl.Categorical,
            "Name": str,
            "Sex": pl.Categorical,
            "SibSp": pl.Categorical,
            "Parch": pl.Categorical,
            "Ticket": str,
            "Cabin": str,
            "Embarked": pl.Categorical,
            "Age": float,
            "Fare": float
        }
        data_frame = pl.read_csv(csv_fp, dtypes=csv_dt)[:100]
        meta_frame = MetaFrame.fit_dataframe(data_frame, spec={"PassengerId": {"unique": True}})
        meta_frame.to_json(json_path)
    return TMP_DIR_PATH

@mark.parametrize("ext", [".csv", ".feather", ".parquet", ".pkl", ".xlsx"])
def test_cli(tmp_dir, ext):
    """A simple integration test for reading and writing using the CLI"""
    
    # create out file path with correct extension
    out_file = tmp_dir / f"titanic{ext}"

    # create command to run in subprocess with arguments
    cmd = [
        Path(sys.executable).resolve(),   # the python executable
        Path("metasynth", "__main__.py"), # the cli script
        "synthesize",                     # the subcommand
        "-n 25",                          # only generate 25 samples
        tmp_dir / "titanic.json",         # the input file
        out_file                          # the output file
    ]

    # Run the cli with different extensions
    result = subprocess.run(cmd, check=False)
    assert result.returncode == 0
    assert out_file.is_file()



import pandas as pd
import polars as pl
import numpy as np
from scipy.stats import poisson

from metasynth.distribution.discrete import UniqueKeyDistribution, DiscreteUniformDistribution,\
    PoissonDistribution
from pytest import mark
from math import fabs


@mark.parametrize(
    "data",
    [
        [1, 2, 3, 4, 5],
        [-399, 12, 1, 492],
        [np.random.randint(0, 1000) for _ in range(1000)],
    ]
)
@mark.parametrize(
    "series_type", [pd.Series, pl.Series]
)
def test_uniform(data, series_type):
    series = series_type(data)
    dist = DiscreteUniformDistribution.fit(series)
    assert dist.low == series.min() and dist.high == series.max()+1
    drawn_values = set([dist.draw() for _ in range(1000)])
    if dist.high - dist.low < 5:
        assert len(drawn_values) == len(series)
    drawn_values = np.array(list(drawn_values))
    assert np.isclose(dist.information_criterion(drawn_values),
                      4+2*len(drawn_values)*(np.log(dist.high-dist.low)))


@mark.parametrize(
    "data,better_than_uniform,consecutive",
    [
        ([1, 2, 3, 4, 5], True, 1),
        ([5, 4, 3, 2, 1], True, 0),
        ([2, 4, 5, 7, 10, 6], True, 0),
        ([-3, 1, -5, 3, -2, 0], True, 0),
        ([-129384, 2198384, 293, 1293840], False, 0),
        ([1, 1, 2, 2, 3, 3], False, 0)
    ]
)
@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_integer_key(data, better_than_uniform, consecutive, series_type):
    series = series_type(data)
    dist = UniqueKeyDistribution.fit(series)
    unif_dist = DiscreteUniformDistribution.fit(series)
    assert dist.low == series.min()
    assert dist.consecutive == consecutive
    assert better_than_uniform == (dist.information_criterion(series) < unif_dist.information_criterion(series))
    assert isinstance(dist, UniqueKeyDistribution)

    drawn_values = np.array([dist.draw() for _ in range(100)])
    if consecutive:
        assert np.all(drawn_values == np.arange(dist.low, dist.low+100))


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_poisson(series_type):
    series = series_type(poisson(mu=10).rvs(1000))
    dist = PoissonDistribution.fit(series)
    dist_unif = DiscreteUniformDistribution.fit(series)
    assert fabs(dist.mu - 10) < 1
    assert dist.information_criterion(series) < dist_unif.information_criterion(series)



from pytest import mark, raises

from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.continuous import UniformDistribution,\
    NormalDistribution
from metasynth.distribution.faker import FakerDistribution
from metasynth.distribution.regex import RegexDistribution
from metasynth.distribution.discrete import DiscreteUniformDistribution
from metasynth.provider import DistributionProviderList


@mark.parametrize(
    "dist_str,dist",
    [
        ("multinoulli", MultinoulliDistribution),
        ("uniform", UniformDistribution),
        ("normal", NormalDistribution),
        ("faker", FakerDistribution),
        ("regex", RegexDistribution),
        ("discrete_uniform", DiscreteUniformDistribution)
    ]
)
def test_util(dist_str, dist):
    provider_list = DistributionProviderList("builtin")
    dist_class = provider_list.find_distribution(dist_str)
    assert dist == dist_class
    if "faker" in dist_str:
        with raises(ValueError):
            provider_list.find_distribution("this is not a distribution")
    new_class = provider_list.find_distribution(dist_class.__name__)
    assert new_class == dist_class



"""Conversion of DataFrames to MetaFrames."""  # pylint: disable=invalid-name

from __future__ import annotations

import json
import pathlib
from copy import deepcopy
from datetime import datetime
from importlib.metadata import version
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import polars as pl

from metasynth.privacy import BasePrivacy, BasicPrivacy
from metasynth.provider import BaseDistributionProvider
from metasynth.validation import validate_gmf_dict
from metasynth.var import MetaVar


class MetaFrame():
    """MetaSynth dataset consisting of variables.

    The MetaSynth dataset structure that is most easily created from
    a polars dataset with the from_dataframe class method.

    Parameters
    ----------
    meta_vars:
        List of variables representing columns in a dataframe.
    n_rows:
        Number of rows in the original dataframe.
    privacy_package:
        Package that supplies the distributions.
    """

    def __init__(self, meta_vars: List[MetaVar],
                 n_rows: Optional[int] = None):
        self.meta_vars = meta_vars
        self.n_rows = n_rows

    @property
    def n_columns(self) -> int:
        """int: Number of columns of the original dataframe."""
        return len(self.meta_vars)

    @classmethod
    def fit_dataframe(
            cls,
            df: pl.DataFrame,
            spec: Optional[dict[str, dict]] = None,
            dist_providers: Union[str, list[str], BaseDistributionProvider,
                                  list[BaseDistributionProvider]] = "builtin",
            privacy: Optional[BasePrivacy] = None):
        """Create a MetaSynth object from a polars (or pandas) dataframe.

        The Polars dataframe should be formatted already with the correct
        datatypes, such as pl.Categorical (or the pandas equivalent).

        Parameters
        ----------
        df:
            Polars dataframe with the correct column dtypes.
        spec:
            Column specifications to modify the defaults. For each of the columns additional
            directives can be supplied here. There are 3 different directives currently supported:

            distribution

            Set the distribution, either with a string that gives one of their
            aliases or an actually fitted BaseDistribution. For example:
            {"distribution": "NormalDistribution"} which is the same as
            {"distribution": NormalDistribution} or
            {"distribution": NormalDistribution(0, 1)}, which are always to set a variable
            to a normal distribution. Note that the first two do not set the parameters
            of the distribution, while the last does.

            unique

            To set a column to be unique/key.
            This is only available for the integer and string datatypes. Setting a variable
            to unique ensures that the synthetic values generated for this variable are unique.
            This is useful for ID or primary key variables, for example. The parameter...
            is ignored when the distribution is set manually. For example:
            {"unique": True}, which sets the variable to be unique or {"unique": False} which
            forces the variable to be not unique. If the uniqueness is not specified, it is
            assumed to be not unique, but might give a warning if it is detected that it might
            be.

            description

            Set the description of a variable: {"description": "Some description."}

            privacy

            Set the privacy level for a variable: {"privacy": DifferentialPrivacy(epsilon=10)}

            prop_missing

            Proportion of missing values for a variable: {"prop_missing": 0.3}

            Any number of the above directives may be set for any number of variables.
        dist_providers:
            Distribution providers to use when fitting distributions to variables.
            Can be a string, provider, or provider type.
        privacy:
            Privacy level to use by default.

        Returns
        -------
        MetaFrame:
            Initialized MetaSynth dataset.
        """
        if privacy is None:
            privacy = BasicPrivacy()
        if spec is None:
            spec = {}
        else:
            spec = deepcopy(spec)

        if set(list(spec)) - set(df.columns):
            raise ValueError(
                "Argument 'spec' includes the specifications for column names that do "
                "not exist in the supplied dataframe:"
                f" '{set(list(spec)) - set(df.columns)}'")
        all_vars = []
        for col_name in df.columns:
            series = df[col_name]
            col_spec = spec.get(col_name, {})
            dist = col_spec.pop("distribution", None)
            unq = col_spec.pop("unique", None)
            description = col_spec.pop("description", None)
            prop_missing = col_spec.pop("prop_missing", None)
            cur_privacy = col_spec.pop("privacy", privacy)
            fit_kwargs = col_spec.pop("fit_kwargs", {})
            if len(col_spec) != 0:
                raise ValueError(
                    f"Unknown spec items '{col_spec}' for variable '{col_name}'.")
            var = MetaVar.detect(
                series,
                description=description,
                prop_missing=prop_missing)
            var.fit(
                dist=dist,
                dist_providers=dist_providers,
                unique=unq,
                privacy=cur_privacy,
                fit_kwargs=fit_kwargs)

            all_vars.append(var)

        return cls(all_vars, len(df))

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary with the properties for recreation."""
        return {
            "n_rows": self.n_rows,
            "n_columns": self.n_columns,
            "provenance": {
                "created by": {
                    "name": "MetaSynth",
                    "version": version("metasynth"),
                },
                "creation time": datetime.now().isoformat()
            },
            "vars": [var.to_dict() for var in self.meta_vars],
        }

    def __getitem__(self, key: Union[int, str]) -> MetaVar:
        """Return meta var either by variable name or index."""
        if isinstance(key, int):
            # If the key is an integer, give the ith variable.
            return self.meta_vars[key]
        if isinstance(key, str):
            # If the key is a string, return the first variable with that name.
            for var in self.meta_vars:
                if var.name == key:
                    return var
            raise KeyError(f"Cannot find variable '{key}'")
        raise TypeError(f"Cannot get item for key '{key}'")

    def __str__(self) -> str:
        """Return an easy to read formatted string for the dataset."""
        vars_formatted = "\n".join(
            f"Column {i + 1}: {str(var)}" for i,
            var in enumerate(
                self.meta_vars))
        return (
            f"# Rows: {self.n_rows}\n"
            f"# Columns: {self.n_columns}\n\n"
            f"{vars_formatted}\n"
        )

    @property
    def descriptions(self) -> dict[str, str]:
        """Return the descriptions of the columns."""
        return {var.name: var.description for var in self.meta_vars
                if var.name is not None and var.description is not None}

    @descriptions.setter
    def descriptions(
            self, new_descriptions: Union[dict[str, str], Sequence[str]]):
        if isinstance(new_descriptions, dict):
            for var_name, new_desc in new_descriptions.items():
                self[var_name].description = new_desc
        else:
            assert len(new_descriptions) == self.n_columns, (
                "Descriptions need to be either a dict or a "
                "sequence with the length of the number of variables.")
            for i_desc, new_desc in enumerate(new_descriptions):
                self[i_desc].description = new_desc

    def export(self, fp: Union[pathlib.Path, str],
               validate: bool = True) -> None:
        """Serialize and export the MetaFrame to a JSON file, following the GMF format.

        Optionally, validate the exported JSON file against the JSON schema(s) included in the
        package.

        Parameters
        ----------
        fp:
            File to write the dataset to.
        validate:
            Validate the JSON file with a schema.
        """
        self_dict = _jsonify(self.to_dict())
        if validate:
            validate_gmf_dict(self_dict)
        with open(fp, "w", encoding="utf-8") as f:
            json.dump(self_dict, f, indent=4)

    def to_json(self, fp: Union[pathlib.Path, str],
                validate: bool = True) -> None:
        """Serialize and export the MetaFrame to a JSON file, following the GMF format.

        This method is a wrapper and simply calls the 'export' function.

        Optionally, validate the exported JSON file against the JSON schema(s) included in the
        package.

        Parameters
        ----------
        fp:
            File to write the dataset to.
        validate:
            Validate the JSON file with a schema.
        """
        self.export(fp, validate)

    @classmethod
    def from_json(cls, fp: Union[pathlib.Path, str],
                  validate: bool = True) -> MetaFrame:
        """Read a MetaFrame from a JSON file.

        Parameters
        ----------
        fp:
            Path to read the data from.
        validate:
            Validate the JSON file with a schema.

        Returns
        -------
        MetaFrame:
            A restored MetaFrame from the file.
        """
        with open(fp, "r", encoding="utf-8") as f:
            self_dict = json.load(f)

        if validate:
            validate_gmf_dict(self_dict)

        n_rows = self_dict["n_rows"]
        meta_vars = [MetaVar.from_dict(d) for d in self_dict["vars"]]
        return cls(meta_vars, n_rows)

    def synthesize(self, n: Optional[int] = None) -> pl.DataFrame:
        """Create a synthetic Polars dataframe.

        Parameters
        ----------
        n:
            Number of rows to generate, if None, use number of rows in original dataframe.

        Returns
        -------
        polars.DataFrame:
            Dataframe with the synthetic data.
        """
        if n is None:
            if self.n_rows is None:
                raise ValueError("Cannot synthesize DataFrame, since number of rows is unknown."
                                 "Please specify the number of rows to synthesize.")
            n = self.n_rows
        synth_dict = {var.name: var.draw_series(n) for var in self.meta_vars}
        return pl.DataFrame(synth_dict)

    def __repr__(self) -> str:
        """Return the MetaFrame as it would be output to JSON."""
        pretty_data = _jsonify(self.to_dict())
        output = json.dumps(pretty_data, indent=4)
        return output


def _jsonify(data):
    if isinstance(data, (list, tuple)):
        return [_jsonify(d) for d in data]
    if isinstance(data, dict):
        return {key: _jsonify(value) for key, value in data.items()}

    if isinstance(data, (np.int8, np.int16, np.int32, np.int64)):
        return int(data)
    if isinstance(data, np.ndarray):
        return _jsonify(data.tolist())
    return data




import polars as pl
from pytest import mark

from metasynth import MetaFrame
from metasynth.distribution.na import NADistribution


@mark.parametrize("dtype", [pl.Float32, pl.Float64, pl.Int32, pl.Int64, pl.Categorical, pl.Utf8])
def test_na(dtype):
    df = pl.DataFrame({"data": pl.Series([None, None, None], dtype=dtype)})
    metadata = MetaFrame.fit_dataframe(df)
    assert isinstance(metadata["data"].distribution, NADistribution)



"""Implemented floating point distributions."""

import numpy as np
from scipy.optimize import minimize
from scipy.stats import expon, lognorm, norm, truncnorm, uniform
from scipy.stats._continuous_distns import FitDataError

from metasynth.distribution.base import metadist, ScipyDistribution


@metadist(implements="core.uniform", var_type="continuous")
class UniformDistribution(ScipyDistribution):
    """Uniform distribution for floating point type.

    This class implements the uniform distribution between a minimum
    and maximum.

    Parameters
    ----------
    min_val: float
        Lower bound for uniform distribution.
    max_val: float
        Upper bound for uniform distribution.
    """

    dist_class = uniform

    def __init__(self, min_val: float, max_val: float):
        self.par = {"min_val": min_val, "max_val": max_val}
        self.dist = uniform(loc=self.min_val, scale=max(self.max_val-self.min_val, 1e-8))

    @classmethod
    def _fit(cls, values):
        return cls(values.min(), values.max())

    def _information_criterion(self, values):
        if np.any(np.array(values) < self.min_val) or np.any(np.array(values) > self.max_val):
            return 2*self.n_par + 100*len(values)
        if np.fabs(self.max_val-self.min_val) < 1e-8:
            return 2*self.n_par - 100*len(values)
        return 2*self.n_par - 2*len(values)*np.log((self.max_val-self.min_val)**-1)

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "min_val": {"type": "number"},
            "max_val": {"type": "number"},
        }


@metadist(implements="core.normal", var_type="continuous")
class NormalDistribution(ScipyDistribution):
    """Normal distribution for floating point type.

    This class implements the normal/gaussian distribution and takes
    the average and standard deviation as initialization input.

    Parameters
    ----------
    mean: float
        Mean of the normal distribution.

    std_dev: float
        Standard deviation of the normal distribution.
    """

    implements = "core.normal"
    dist_class = norm

    def __init__(self, mean: float, std_dev: float):
        self.par = {"mean": mean, "std_dev": std_dev}
        self.dist = norm(loc=mean, scale=max(std_dev, 1e-8))

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mean": {"type": "number"},
            "std_dev": {"type": "number"},
        }


@metadist(implements="core.lognormal", var_type="continuous")
class LogNormalDistribution(ScipyDistribution):
    """Log-normal distribution for floating point type.

    This class implements the log-normal mu and sigma as initialization input.

    Parameters
    ----------
    sigma: float
        Controls the width of the distribution.
    mu: float
        Controls the mean of the distribution.
    """

    dist_class = lognorm

    def __init__(self, mu: float, sigma: float):  # pylint: disable=invalid-name
        self.par = {"mu": mu, "sigma": sigma}
        self.dist = lognorm(s=max(sigma, 1e-8), scale=np.exp(mu))

    @classmethod
    def _fit(cls, values):
        try:
            sigma, _, scale = cls.dist_class.fit(values, floc=0)
        except FitDataError:
            return cls(0, 1)
        return cls(np.log(scale), sigma)

    @classmethod
    def default_distribution(cls):
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mu": {"type": "number"},
            "sigma": {"type": "number"},
        }


@metadist(implements="core.truncated_normal", var_type="continuous")
class TruncatedNormalDistribution(ScipyDistribution):
    """Truncated normal distribution for floating point type.

    Parameters
    ----------
    lower_bound: float
        Lower bound of the truncated normal distribution.
    upper_bound: float
        Upper bound of the truncated normal distribution.
    mu: float
        Mean of the non-truncated normal distribution.
    sigma: float
        Standard deviation of the non-truncated normal distribution.
    """

    dist_class = truncnorm

    def __init__(self, lower_bound: float, upper_bound: float,
                 mu: float, sigma: float):
        self.par = {"lower_bound": lower_bound, "upper_bound": upper_bound,
                    "mu": mu, "sigma": sigma}
        a, b = (lower_bound-mu)/sigma, (upper_bound-mu)/sigma
        self.dist = truncnorm(a=a, b=b, loc=mu, scale=max(sigma, 1e-8))

    @classmethod
    def _fit(cls, values):
        lower_bound = values.min() - 1e-8
        upper_bound = values.max() + 1e-8
        return cls._fit_with_bounds(values, lower_bound, upper_bound)

    @classmethod
    def _fit_with_bounds(cls, values, lower_bound, upper_bound):
        def minimizer(param):
            mu, sigma = param
            a, b = (lower_bound-mu)/sigma, (upper_bound-mu)/sigma
            dist = truncnorm(a=a, b=b, loc=mu, scale=sigma)
            return -np.sum(dist.logpdf(values))

        x_start = [(lower_bound+upper_bound)/2, (upper_bound-lower_bound)/4]
        mu, sigma = minimize(minimizer, x_start,
                             bounds=[(None, None),
                                     ((upper_bound-lower_bound)/100, None)]).x
        return cls(lower_bound, upper_bound, mu, sigma)

    @classmethod
    def default_distribution(cls):
        return cls(-1, 2, 0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "lower_bound": {"type": "number"},
            "upper_bound": {"type": "number"},
            "mu": {"type": "number"},
            "sigma": {"type": "number"},
        }


@metadist(implements="core.exponential", var_type="continuous")
class ExponentialDistribution(ScipyDistribution):
    """Exponential distribution for floating point type.

    This class implements the exponential distribution with the rate as its
    single parameter.

    Parameters
    ----------
    rate: float
        Rate of the exponential distribution. This is equal to 1/mean of the distribution.
    """

    dist_class = expon

    def __init__(self, rate: float):
        self.par = {"rate": rate}
        self.dist = expon(loc=0, scale=1/max(rate, 1e-8))

    @classmethod
    def _fit(cls, values):
        values = values.filter(values > 0)
        if len(values) == 0:
            return cls.default_distribution()
        return cls(rate=1/expon.fit(values, floc=0)[1])

    @classmethod
    def default_distribution(cls):
        return cls(1.0)

    @classmethod
    def _param_schema(cls):
        return {
            "rate": {"type": "number"}
        }



"""Load/create different demo datasets."""

from pathlib import Path
import random

try:
    from importlib_resources import files
except ImportError:
    from importlib.resources import files  # type: ignore

import pandas as pd
import numpy as np
import wget

from metasynth.distribution.datetime import UniformDateTimeDistribution, UniformTimeDistribution
from metasynth.distribution.datetime import UniformDateDistribution


def create_titanic_demo(output_fp: Path) -> Path:
    """Create demo dataset for the titanic dataset.

    Arguments
    ---------
    output_fp:
        File to write the demonstration table to.

    Returns
    -------
        Output file location.
    """
    titanic_fp = Path("titanic.csv")
    if output_fp.is_file():
        return output_fp
    if not titanic_fp.is_file():
        wget.download(
            "https://raw.githubusercontent.com/pandas-dev/pandas/main/doc/data/titanic.csv")
    dframe = pd.read_csv(titanic_fp)
    np.random.seed(1283742)
    random.seed(1928374)

    # Convert Age to a nullable integer.
    dframe["Age"] = dframe["Age"].round().astype("Int64")

    # Add a date column.
    date_dist = UniformDateDistribution.default_distribution()
    dframe["Birthday"] = [date_dist.draw() if np.random.rand() < 0.9 else pd.NA
                          for _ in range(len(dframe))]

    # Add a time column.

    time_dist = UniformTimeDistribution.default_distribution()
    dframe["Board time"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA
                            for _ in range(len(dframe))]

    # Add a datetime column
    time_dist = UniformDateTimeDistribution.default_distribution()
    dframe["Married since"] = [time_dist.draw() if np.random.rand() < 0.9 else pd.NA
                               for _ in range(len(dframe))]

    dframe["all_NA"] = [pd.NA for _ in range(len(dframe))]
    # Remove some columns for brevity and write to a file.
    dframe = dframe.drop(["SibSp", "Pclass", "Survived"], axis=1)
    dframe.to_csv(output_fp, index=False)
    return output_fp


def demo_file(name: str = "titanic") -> Path:
    """Get the path for a demo data file.

    Arguments
    ---------
    name:
        Name of the demo dataset.

    Returns
    -------
        Path to the dataset.
    """
    file_name = None
    if name == "titanic":
        file_name = "demo_titanic.csv"
    if file_name is None:
        raise ValueError(f"No demonstration dataset with name '{name}'")

    return files(__package__) / file_name
